@@ -49,24 +49,50 @@ proc Aliases_Pref { {nalias {}} } {
 	.aliaspref.addr.listbox config -exportselection 0
 
 	# Create search bindings in the entries.
-	# Need to suck up the class bindings for Entry first
-	foreach seq [bind Entry] {
-	    bind .aliaspref.alias.entry $seq [bind Entry $seq]
-	    bind .aliaspref.addr.entry $seq [bind Entry $seq]
-	}
-	bind .aliaspref.alias.entry <Any-Tab> { focus .aliaspref.addr.entry }
-	bind .aliaspref.alias.entry <Any-Key> \
-	    "[bind Entry <Any-Key>] ; Widget_ListSearch .aliaspref.alias"
-	bind .aliaspref.alias.entry <Any-space> \
-	    {set alias [AliasesComplete .aliaspref.alias]
-	     Aliases_Insert}
-	bind .aliaspref.alias.entry <Control-c> \
-	    {Aliases_Delete}
-	bind .aliaspref.addr.entry <Any-Tab> { focus .aliaspref.alias.entry }
-	bind .aliaspref.addr.entry <Any-Key>\
-	    "[bind Entry <Any-Key>] ; Widget_ListSearch .aliaspref.addr"
-	bind .aliaspref.addr.entry <Control-space> {
-	    set address [AliasesComplete .aliaspref.addr]
+	global tk_version
+	if {$tk_version < 4.0} {
+	    # Need to suck up the class bindings for Entry first
+	    foreach seq [bind Entry] {
+		bind .aliaspref.alias.entry $seq [bind Entry $seq]
+		bind .aliaspref.addr.entry $seq [bind Entry $seq]
+	    }
+	    bind .aliaspref.alias.entry <Any-Tab> { focus .aliaspref.addr.entry }
+	    bind .aliaspref.alias.entry <Any-Key> \
+		"[bind Entry <Any-Key>] ; Widget_ListSearch .aliaspref.alias"
+	    bind .aliaspref.alias.entry <Any-space> \
+		{set alias [AliasesComplete .aliaspref.alias]
+		 Aliases_Insert}
+	    bind .aliaspref.alias.entry <Control-c> \
+		{Aliases_Delete}
+	    bind .aliaspref.addr.entry <Any-Tab> { focus .aliaspref.alias.entry }
+	    bind .aliaspref.addr.entry <Any-Key>\
+		"[bind Entry <Any-Key>] ; Widget_ListSearch .aliaspref.addr"
+	    bind .aliaspref.addr.entry <Control-space> {
+		set address [AliasesComplete .aliaspref.addr]
+	    }
+	} else {
+	    bindtags .aliaspref.alias.entry \
+		[list .aliaspref.alias.entry Entry SearchAlias]
+	    bind SearchAlias <KeyPress> \
+		 {Widget_ListSearch .aliaspref.alias}
+	    bind .aliaspref.alias.entry <space> \
+		{set alias [AliasesComplete .aliaspref.alias]
+		 Aliases_Insert
+		 break}
+	    bind .aliaspref.alias.entry <Control-c> \
+		{Aliases_Delete ; break}
+
+	    bindtags .aliaspref.addr.entry \
+		[list .aliaspref.addr.entry Entry SearchAddress]
+	    bind SearchAddress <KeyPress> \
+		 {Widget_ListSearch .aliaspref.addr}
+	    bind .aliaspref.addr.entry <Control-space> {
+		{set address [AliasesComplete .aliaspref.addr]
+		 break}
+	    }
+	    foreach e {.aliaspref.alias.entry .aliaspref.addr.entry} {
+		Exmh_Debug bindtags $e [bindtags $e]
+	    }
 	}
 	wm minsize .aliaspref 1 1
 
@@ -84,53 +110,13 @@ proc AliasesComplete { parent } {
     Exmh_Status AliasesComplete
     set i [$parent.listbox curselection]
     if {$i != {}} {
-	return [$parent.listbox get $i]
+	return [$parent.listbox get [lindex $i 0]]
     } else {
 	return [$parent.entry get]
     }
 }
 proc Aliases_Help {} {
-    global mhProfile
-    if [Exwin_Toplevel .aliashelp "Aliases Help" Help] {
-	Widget_Message .aliashelp msg -aspect 1000 -text \
-"Aliases are short names for email addresses.
-You can have multiple address as part of one alias.
-Aliases can refer to other aliases that are defined
-before them.
-
-The email address can include rfc822 comments in parenthesis, like:
-(Brent Welch) welch@parc.xerox.com
-
-BUG WARNING: the addresses are not really parsed,
-and if you have a comma inside a comment, it screws up.
-Don't use this browser in that case.
-
-Use Import to import aliases from .mailrc files.
-
-Search lists by typing in the entry below them.
-Hit space when the correct alias is selected.
-(Use <Control-space> when the correct address is selected.
-That'll work for alias completion, too.)
-
-<space> will define a new alias if what you have typed
-does not match an existing alias.
-<Control-c> will delete the matching alias.
-
-You can select and paste out of the Alias list.
-The selection is not exported from the Address list.
-This asymmetry means you can select an address,
-then select and paste another version of that address
-into the entry widget below the address list,
-and (finally) click Change to update that address.
-
-Your aliases are stored in [Mh_Pathname $mhProfile(aliasfile)]
-
-Exmh remembers the placement of the window, but not necessarily
-its size.  For best results, play with these resources
-*Alias*alias.listbox.geometry:  20x10
-*Alias*addr.listbox.geometry:	30x10
-"
-    }
+    Help Aliases "Tips for the Aliases interface"
 }
 
 #
@@ -140,6 +126,9 @@ its size.  For best results, play with these resources
 #
 proc AliasesInitProfile {} {
     global mhProfile aliasOK
+    if [winfo exists .aliasfile] {
+	return
+    }
     Widget_Toplevel .aliasfile "Setup Alias File"
     Widget_Message .aliasfile msg -aspect 1000 -text "
 The MH aliases are kept in a file (you choose the name)
@@ -204,12 +193,12 @@ proc Aliases_Rebuild {{item {}}} {
     set names [AliasesList]
     eval .aliaspref.alias.listbox insert end $names
     if {$item == {}} {
-	.aliaspref.alias.listbox select clear
+	Widget_ListboxClear .aliaspref.alias.listbox
 	Widget_ListboxYview .aliaspref.alias.listbox $top
 	set alias {}
     } else {
 	set index [lsearch -exact $names $item]
-	.aliaspref.alias.listbox select from $index
+	Widget_ListboxSelect .aliaspref.alias.listbox  $index
 	Widget_ListboxYview .aliaspref.alias.listbox $index
 	set alias $item
     }
@@ -232,11 +221,6 @@ proc AliasesList {} {
     }
     return $res
 }
-proc Widget_ListboxYview { list i } {
-    set height [lindex [split [lindex [$list config -geometry] 4] x] 1]
-    $list yview [expr $i - $height/2]
-}
-
 #
 # Address_Rebuild:
 #
@@ -248,7 +232,7 @@ proc Address_Rebuild {{item {}}} {
     global aliases address alias
 
     # rebuild the address list based on the current alias selected
-    if {[string length $alias] > 0} {
+    if {([string length $alias] > 0) && [info exists aliases($alias)]} {
 	.aliaspref.addr.listbox delete 0 end
 	set names [lsort $aliases($alias)]
 	eval .aliaspref.addr.listbox insert end $names
@@ -259,10 +243,10 @@ proc Address_Rebuild {{item {}}} {
 
     # now select appropriate item from the list
     if {$item == {}} {
-	.aliaspref.addr.listbox select clear
+	Widget_ListboxClear .aliaspref.addr.listbox
     } else {
 	set index [lsearch -exact $names $item]
-	.aliaspref.addr.listbox select from $index
+	Widget_ListboxSelect .aliaspref.addr.listbox $index
 	Widget_ListboxYview .aliaspref.addr.listbox $index
 	set address [.aliaspref.addr.listbox get $index]
     }
@@ -339,7 +323,7 @@ proc Aliases_Delete {} {
 	if {$ix >= 0} {
 	    AliasesDirty
 	    set aliasesOrder [lreplace $aliasesOrder $ix $ix]
-	    unset aliases($alias)
+	    catch {unset aliases($alias)}
 	    Aliases_Rebuild
 	}
     }
@@ -399,7 +383,9 @@ proc Address_Change {} {
 #
 proc Address_Delete {} {
     global aliases alias address
-    if {([string length $address] > 0) && ([string length $alias] > 0)} {
+    if {([string length $address] > 0) && 
+	([string length $alias] > 0) &&
+	[info exists aliases($alias)]} {
 	set pos [lsearch $aliases($alias) $address]
 	if {$pos >= 0} {
 	    set aliases($alias) \
@@ -513,7 +499,7 @@ proc Aliases_Load {} {
 	    set aliases($key) \
 		[concat $aliases($key) [split [string trim $value "\t ,"] ,]]
 	} else {
-	    if [regexp {^[ 	]*[;#]} $line] {
+	    if [regexp {^[ 	]*[;#:]} $line] {
 		lappend aliasesOrder $line
 	    } elseif [regexp "^\[ \t\]*$" $line] {
 		lappend aliasesOrder $line
@@ -581,7 +567,7 @@ proc Aliases_Save {} {
 
     if [info exists aliases] {
 	if [catch {open [Mh_Pathname $mhProfile(aliasfile)] w} output] {
-	    puts stderr "Cannot write [Mh_Pathname $mhProfile(aliasfile)]: $input"
+	    puts stderr "Cannot write [Mh_Pathname $mhProfile(aliasfile)]: $output"
 	    return
 	}
 
@@ -592,7 +578,7 @@ proc Aliases_Save {} {
 		puts $output $alias
 	    } elseif [regexp {^[ 	]*<} $alias] {
 		puts $output $alias
-	    } else {
+	    } elseif [info exists aliases($alias)] {
 		if [catch {set aliases_sep($alias)} sep] {
 		    set sep :
 		}
