@@ -12,7 +12,20 @@
 # any specification.
 
 proc Fdisp_Init {} {
-    global fdisp mhProfile
+    global fdisp mhProfile exmh flist
+
+    if {[info exists exmh(newuser)] && [info exists flist(allfolders)]} {
+	set N [llength $flist(allfolders)]
+	if {$N < 5} {
+	    set fdisp(maxLines) 1
+	} elseif {$N < 15} {
+	    set fdisp(maxLines) 2
+	} elseif {$N < 30} {
+	    set fdisp(maxLines) 3
+	} else {
+	    set fdisp(maxLines) 4
+	}
+    }
 
     Preferences_Add "Folder Display" \
 "These items affect the display of the labels in the folder display window.
@@ -72,6 +85,10 @@ be implemented by triggering on <Enter> to other labels." }
 }
 proc FdispFixupMaxLines { args } {
     global exwin fdisp
+    if [catch {expr {$fdisp(maxLines) * 2}}] {
+	set fdisp(maxLines) $fdisp(lastMaxLines)
+	return	;# bogus value
+    }
     if {$fdisp(maxLines) != $fdisp(lastMaxLines)} {
 	set fdisp(width,canvas) 0
 	set fdisp(maxLines,$fdisp(canvas)) $fdisp(maxLines)
@@ -96,12 +113,15 @@ proc FdispFixupToplevel { args } {
     set fdisp(lastToplevel) $fdisp(toplevel)
 }
 proc FdispMakeToplevel { } {
-    global fdisp
+    global fdisp tk_version
     Exwin_Toplevel .fl "Folder list" Fltop nomenu
     set fdisp(topWidget) .fl
     wm minsize $fdisp(topWidget) 100 30
     wm protocol .fl WM_DELETE_WINDOW FdispDeleted
     FdispMakeCanvas $fdisp(topWidget)
+    if {$tk_version >= 4.0} {
+	focus $fdisp(canvas)
+    }
     set icon [option get $fdisp(topWidget) iconposition IconPosition]
     catch {
 	Exwin_IconPosition $fdisp(topWidget) $icon
@@ -117,10 +137,17 @@ proc FdispMakeToplevel { } {
 proc Fdisp_Checkpoint { varName } {
     # Add Xresources lines to $varName that save window size
     upvar $varName newstuff
-    global fdisp
+    global fdisp tk_version
     catch {
-	set width [winfo width $fdisp(topWidget).can]
-	set height [winfo height $fdisp(topWidget).can]
+	set can $fdisp(topWidget).can
+	set width [winfo width $can]
+	set height [winfo height $can]
+	set bd [lindex [$can config -borderwidth] 4]
+	if {$tk_version >= 4.0} {
+	    incr bd [$can cget -highlightthickness]
+	}
+	set width [expr $width - 2*$bd]
+	set height [expr $height - 2*$bd]
 	lappend newstuff "*Fltop.Canvas.height:\t$height"
 	lappend newstuff "*Fltop.Canvas.width:\t$width"
     }
@@ -139,6 +166,36 @@ proc FdispMakeCanvas { frame } {
     set fdisp(canvas) [canvas $frame.can -bd 2 -relief raised ]
     set s [scrollbar $frame.sv -command [list $fdisp(canvas) yview]]
     $fdisp(canvas) configure -yscrollcommand [list $s set]
+
+    # Find out how big labels are
+    if [catch {
+	set id [$fdisp(canvas) create text 0 0 \
+	    -anchor nw -justify center -text 0123456789 -font $fdisp(font)]
+    } err] {
+	Exmh_Status $err
+	set fdisp(font) fixed
+	set id [$fdisp(canvas) create text 0 0 \
+	    -anchor nw -justify center -text 0123456789 -font $fdisp(font)]
+    }
+    set size [$fdisp(canvas) bbox $id]
+    set fdisp(itemHeight) [expr {[lindex $size 3] - [lindex $size 1]}]
+    set fdisp(charWidth) [expr {([lindex $size 2] - [lindex $size 0])/10}]
+    $fdisp(canvas) delete $id
+
+    if {$tk_version < 4.0} {
+	$fdisp(canvas) configure -scrollincrement \
+	    [expr {$fdisp(itemHeight)+$fdisp(ygap)+1}]
+    } else {
+	catch {
+	    $fdisp(canvas) configure -yscrollincrement \
+		[expr {$fdisp(itemHeight)+$fdisp(ygap)+1}]
+	}
+	bindtags $fdisp(canvas) [list $fdisp(canvas) Command [winfo toplevel $fdisp(canvas)] all]
+    }
+    if {! $fdisp(toplevel)} {
+	set h [expr {$fdisp(maxLines)*($fdisp(itemHeight) + $fdisp(ygap)) + $fdisp(ygap)}]
+	$fdisp(canvas) configure -height $h
+    }
     bind $fdisp(canvas) <2> {%W scan mark %x %y}
     bind $fdisp(canvas) <B2-Motion> {%W scan dragto %x %y}
     if {! $fdisp(toplevel)} {
@@ -148,27 +205,30 @@ proc FdispMakeCanvas { frame } {
 	}
     }
     bind $fdisp(canvas) <Configure> FdispCanvasConfigure
-    if {$tk_version >= 3.3} {
-	pack $s -side $exwin(scrollbarSide) -fill y
-	pack $fdisp(canvas) -side $exwin(scrollbarSide) -fill both -expand 1
-    } else {
-	pack append $frame $s [list $exwin(scrollbarSide) fill] \
-		    $fdisp(canvas) [list $exwin(scrollbarSide) expand fill]
-    }
+    pack $s -side $exwin(scrollbarSide) -fill y
+    pack $fdisp(canvas) -side $exwin(scrollbarSide) -fill both -expand 1
     # fdisp popup color hack
     if {[tk colormodel $fdisp(canvas)] != "color"} {
 	if {! [regexp {black|white} $fdisp(c_popup)]} {
 	    set fdisp(c_popup) [lindex [$fdisp(canvas) config -bg] 4]
 	}
     }
+    FdispDragAttach canvas
 }
 proc FdispFixupFont { args } {
     global exwin fdisp
     if {$fdisp(lastFont) != $fdisp(font)} {
 
 	# Find out how big labels are
-	set id [$fdisp(canvas) create text 0 0 \
+	if [catch {
+	    set id [$fdisp(canvas) create text 0 0 \
 		-anchor nw -justify center -text foo -font $fdisp(font)]
+	} err] {
+	    Exmh_Status $err
+	    set fdisp(font) fixed
+	    set id [$fdisp(canvas) create text 0 0 \
+		-anchor nw -justify center -text foo -font $fdisp(font)]
+	}
 	set size [$fdisp(canvas) bbox $id]
 	set fdisp(itemHeight) [expr {[lindex $size 3] - [lindex $size 1]}]
 	$fdisp(canvas) delete $id
@@ -201,23 +261,11 @@ proc Fdisp_Window { parent } {
 	FdispMakeFrame
      }
 
-    # Find out how big labels are
-    set id [$fdisp(canvas) create text 0 0 \
-	    -anchor nw -justify center -text 0123456789 -font $fdisp(font)]
-    set size [$fdisp(canvas) bbox $id]
-    set fdisp(itemHeight) [expr {[lindex $size 3] - [lindex $size 1]}]
-    set fdisp(charWidth) [expr {([lindex $size 2] - [lindex $size 0])/10}]
-    $fdisp(canvas) delete $id
-    $fdisp(canvas) configure -scrollincrement \
-	    [expr {$fdisp(itemHeight)+$fdisp(ygap)+1}]
-    if {! $fdisp(toplevel)} {
-	set h [expr {$fdisp(maxLines)*($fdisp(itemHeight) + $fdisp(ygap)) + $fdisp(ygap)}]
-	$fdisp(canvas) configure -height $h
-    }
 
     global fcache
     if $fcache(enabled) {
 	Fcache_CreateWindow
+	FdispDragAttach cache
     }
 
     set fdisp(folder) .
@@ -320,7 +368,7 @@ proc FdispLayoutInner { can x1 y1 width folderSet folder bindProc {skipSelf 0} {
 	    }
 	} else {
 	    if {$can == "cache"} {
-		set text $f
+		set text [Fcache_FolderName $f]
 	    } else {
 		set text [file tail $f]
 	    }
@@ -570,13 +618,8 @@ proc FdispWhichLabel { can f } {
     # Figure out what label to highlight, handling nesting
     global fdisp mhProfile
 
-    set map [FdispGetMap $can $f]
-    if {$map != {}} {
-	if {[string compare $f ".."] == 0} {
-	    return {}
-	} else {
-	    return $f
-	}
+    if [FdispNotDotDot $can $f] {
+	return $f
     }
     while {[string compare $f "."] && [string compare $f "/"]} {
 	set nf [file dirname $f]
@@ -584,13 +627,8 @@ proc FdispWhichLabel { can f } {
 	    break
 	}
 	set f $nf
-	set map [FdispGetMap $can $f]
-	if {$map != {}} {
-	    if {[string compare $f ".."] == 0} {
-		return {}
-	    } else {
-		return $f
-	    }
+	if [FdispNotDotDot $can $f] {
+	    return $f
 	}
     }
     return {}
@@ -601,8 +639,7 @@ proc FdispAllLabels { can f } {
     global fdisp mhProfile
 
     set res {}
-    set map [FdispGetMap $can $f]
-    if {($map != {}) && [string compare $f ".."]} {
+    if [FdispNotDotDot $can $f] {
 	lappend res $f
     }
     while {[string compare $f "."] && [string compare $f "/"]} {
@@ -611,13 +648,29 @@ proc FdispAllLabels { can f } {
 	    break
 	}
 	set f $nf
-	set map [FdispGetMap $can $f]
-	if {($map != {}) && [string compare f ".."]} {
+	if [FdispNotDotDot $can $f] {
 	    lappend res $f
 	}
     }
     return $res
 }
+# See if the folder label displayed for $f is ".." (and is displayed at all)
+proc FdispNotDotDot { can f } {
+    global fdisp
+    set map [FdispGetMap $can $f]
+    if {$map != {}} {
+	if [catch {lindex [$fdisp($can) itemconfigure $map -text] 4} l] {
+	    if [string compare $f ".."] {
+		return 1
+	    }
+	} else {
+	    if [string compare $l ".."] {
+		return 1
+	    }
+	}
+    }
+    return 0
+}
 proc Fdisp_HighlightCur { f } {
     global fdisp
 
@@ -689,7 +742,11 @@ proc Fdisp_HighlightUnseen { f } {
 }
 proc FdispHighlightUnseen { can f } {
     global exmh fdisp
-    set ll [FdispAllLabels $can $f]
+    if {$can != "cache"} {
+	set ll [FdispAllLabels $can $f]
+    } else {
+	set ll [list $f]
+    }
     set canvas $fdisp($can)
     foreach l $ll {
 	set id [FdispGetMap $can $l]
@@ -715,22 +772,18 @@ proc Fdisp_UnHighlightUnseen { f } {
 proc FdispUnHighlightUnseen { canvas can f } {
     global exmh fdisp
     set ll [FdispAllLabels $can $f]
-    Exmh_Debug FdispUnHighlightUnseen $can $f $ll
     set canvas $fdisp($can)
     foreach l $ll {
 	set id [FdispGetMap $can $l]
 	set box [FdispGetBmap $can $l]
 	set stillLight 0
-	Exmh_Debug Tags $l [$canvas gettags $id]
 	foreach tag [$canvas gettags $id] {
 	    if [string match leaf=* $tag] {
 		set leaf [lindex [split $tag =] 1]
 		if {[string compare $leaf $f] == 0} {
 		    $canvas dtag $id $tag
-		    Exmh_Debug Dtag
 		} else {
 		    set stillLight 1
-		    Exmh_Debug StillLight
 		}
 	    }
 	}
@@ -746,7 +799,8 @@ proc Fdisp_Lines { canvas labels } {
     set x $fdisp(xgap)
     set lines 1
     set width [winfo width $canvas]
-    foreach f $labels {
+    foreach folder $labels {
+	set f [Fcache_FolderName $folder]
 	set id [Fdisp_Label $canvas $f $f]
 	set bbox [$canvas bbox $id]
 	set twidth [expr [lindex $bbox 2]-[lindex $bbox 0]]
@@ -759,3 +813,120 @@ proc Fdisp_Lines { canvas labels } {
     }
     return $lines
 }
+
+#
+# Interface to Drag & Drop
+#
+set fdispDrag(callback) FdispDragRelease
+set fdispDrag(types) {folder filename}
+set fdispDrag(formats) string
+set fdispDrag(format,folder) string
+set fdispDrag(format,filename) string
+set fdispDrag(type,string) folder
+set fdispDrag(decorate) FdispDragWindow
+
+proc FdispDragAttach {where} {
+	global fdisp
+
+	Drag_Attach $fdisp($where) FdispDragSelect Shift $fdisp(navbutton)
+	if [string match cache $where] {
+		Drop_Attach $fdisp(cache) FdispDropCache
+	} else {
+		Drop_Attach $fdisp(canvas) FdispDropCanvas
+	}
+}
+
+# A drag was dropped on the cache
+proc FdispDropCache {w args} {
+	global dragging
+
+	if ![info exists dragging(data,folder)] return
+	set folder $dragging(data,folder)
+
+	# Add the folder to the cache
+	Fcache_Folder $folder
+}
+
+# A drag was dropped on the canvas
+proc FdispDropCanvas {w args} {
+	global fdisp dragging
+
+	if ![info exists dragging(data,folder)] return
+	set folder $dragging(data,folder)
+
+	# If dropped on the folder display and source was cache,
+	# remove the folder from the cache
+	if {[info exists fdisp(cache)] && 
+	    $dragging(source) == $fdisp(cache)} {
+		Fcache_FolderDiscard $folder
+	}
+}
+
+# Called when after a drag we sourced has been dropped
+proc FdispDragRelease {dstw args} {
+
+	global fdisp dragging
+	set folder $dragging(data,folder)
+
+tlog-add .t "released on $dstw"
+
+	# If we tossed it somewhere unknown, Add the folder to the cache
+	if {$dragging(source) == $fdisp(canvas) && $dstw != $fdisp(canvas) &&
+	    "$dstw" != {}} {
+		Fcache_Folder $folder
+	}
+}
+
+# Drag Selected
+proc FdispDragSelect {c x y wx wy} {
+	global fdisp
+
+	set wy [$c canvasy $wy]
+	set items [$c find overlapping $wx $wy $wx $wy]
+
+	# Find what folder we're over
+	foreach item $items {
+		set tags [$c gettags $item]
+		set which [lsearch -glob $tags F*=*]
+		if {$which >= 0} {
+			set tag [lindex $tags $which]
+			regsub -- .*=(.*) $tag {\1} folder
+			break
+		}
+	}
+	if ![info exists folder] return
+
+	# Hand off to Drag code
+	global fdispDrag mhProfile
+	set fdispDrag(source) $c
+	set fdispDrag(data,folder) $folder
+	set fdispDrag(data,filename) $mhProfile(path)/$folder
+
+	Drag_Source fdispDrag $x $y
+
+}
+
+# How do decorate the Drag window
+proc FdispDragWindow {w} {
+	global fdisp dragging
+
+	set c $w.fdisp
+	if ![winfo exists $c] {
+		set height [expr $fdisp(itemHeight) + $fdisp(ygap)]
+		canvas $c -height $height
+	}
+
+	pack $c
+	catch {$c delete all}
+
+	set f $dragging(data,folder)
+
+	set id [Fdisp_Label $c $f $f]
+	set bbox [$c bbox $id]
+	set twidth [expr [lindex $bbox 2]-[lindex $bbox 0]]
+	set theight [expr [lindex $bbox 3]-[lindex $bbox 1]]
+	$c move $id [expr $fdisp(xgap)/2 + 1] [expr $fdisp(ygap)/2]
+	set width [expr $twidth + $fdisp(xgap)]
+	$c config -width $width
+	set bid [Fdisp_Box $c $id leaf {}]
+}
