@@ -1,9 +1,12 @@
 #
 # flist.tcl
 #
-# Manage the folder list.  This includes information about which folders
-# have nested folders under them, and which folders have unseen messages.
-# The display of this information is done by the fdisp.tcl module
+# Manage the folder list.
+# For folder display (fdisp.tcl):
+#	What folders have nested folders under them
+#	What folders have unseen messages
+# For scan listing (ftoc.tcl):
+#	What messages are unread.
 #
 # Some of the routines here are set up to run in a background interpreter.
 # When you see calls to BgRPC it is invoking the routine in the
@@ -21,34 +24,91 @@
 
 proc Flist_Init {} {
     global flist
-    FlistInitVars		;# Reset unseen msgs state
+    FlistResetVars		;# Reset unseen msgs state
     set flist(context) {}
     set flist(contextMtime) 0
     set flist(cacheFileMtime) 0
     Flist_FindAllFolders
 }
-proc FlistInitVars {} {
+proc FlistResetVars {} {
     global flist
     set flist(unseen) {}	;# Sequence of folders to visit
     set flist(unvisited) {}	;# Unseen folders not yet visited
-    set flist(newMsgs) 0
+    Exmh_Debug FlistInitVars
+    set flist(unvisitedNext) {}	;# Temporary copy (next iteration)
+    set flist(newMsgs) 0	;# Total count of unseen messages
     set flist(newFolders) 0
+
+    # flist(new,$folder)	is the number of unseen in a folder
+    # flist(newseq,$folder)	is the message id's of unseen messages
+    # flist(mtime,$folder)	is modify time of .mh_sequences file
+    foreach x [array names flist] {
+	if [regexp {^(new|newseq|mtime),} $x] {
+	    # reset per-folder new and newseq state
+	    unset flist($x)
+	}
+    }
+    if ![info exists flist(debug)] {
+	set flist(debug) 0
+    }
+    if {$flist(debug)} {
+	trace variable flist(newMsgs) w FlistTraceNewMsgs
+	trace variable flist(unseen) w FlistTraceUnseen
+	trace variable flist(unvisited) w FlistTraceUnvisited
+	set flist(listbox) .flistbox
+	set f $flist(listbox)
+	if ![winfo exists $f] {
+	    Exwin_Toplevel $f "Flist Debug" Flist
+	    Widget_Label $f newMsgs {top fillx } -textvariable flist(newMsgs)
+	    Widget_Frame $f top Labels
+	    Widget_Label $f.top unseen {left fill expand} -text Unseen
+	    Widget_Label $f.top unvisited {left fill expand} -text Unvisited
+	    FontWidget listbox $f.unseen
+	    FontWidget listbox $f.unvisited
+	    pack $f.unseen $f.unvisited -side left -fill both -expand true
+	}
+    }
+}
+proc FlistTraceNewMsgs {args} {
+    global flist
+    set l [info level]
+    incr l -1
+    Exmh_Debug flist(newMsgs) => $flist(newMsgs) : [info level $l] $args
+}
+proc FlistTraceUnseen {args} {
+    global flist
+    $flist(listbox).unseen delete 0 end
+    foreach f $flist(unseen) {
+	$flist(listbox).unseen insert end $f
+    }
+}
+proc FlistTraceUnvisited {args} {
+    global flist
+    $flist(listbox).unvisited delete 0 end
+    foreach f $flist(unvisited) {
+	$flist(listbox).unvisited insert end $f
+    }
 }
 
 ### Routines to find all folders, figure out which have nested folders, etc.
 
 proc Flist_Refresh {} {
     global flist
+    FlistResetVars
     FlistFindAllInner
     Fdisp_Redisplay
-    Flist_FindUnseen
+    Flist_FindUnseen 1
+    Inc_PresortFinish
 }
 
 proc Flist_FindAllFolders {} {
     global flist mhProfile flistSubCache flistParents
 
-    set flist(cacheFile) $mhProfile(path)/.folders
-    if ![file readable $flist(cacheFile)]  {
+    if ![info exists flist(cacheFile)] {
+	set flist(cacheFile) $mhProfile(path)/.folders
+    }
+    if {![file readable $flist(cacheFile)] ||
+	[file size $flist(cacheFile)] == 0} {
 	FlistFindAllInner
     } else {
 	if {![info exists flist(allfolders)]||
@@ -58,6 +118,7 @@ proc Flist_FindAllFolders {} {
 	    close $in
 	    set flist(cacheFileMtime) [file mtime $flist(cacheFile)]
 	    FlistSubFoldersInit
+	    BgAction FlistUnseenFoldersInit FlistUnseenFoldersInit
 	}
     }
 
@@ -67,10 +128,7 @@ proc FlistFindAllInner {} {
     Widget_Toplevel .scanning "Scanning..."
     Widget_Message .scanning msg -cursor watch -text "
 Scanning for nested folders.
-(folders -fast -recurse)
-This can be very slow if you have
-not patched the MH folders command.
-A patch comes with exmh.
+(folders -all -fast -recurse)
 
 The results are cached in
 $mhProfile(path)/.folders
@@ -82,7 +140,7 @@ Please wait...
 "
     Exmh_Status "Scanning for nested folders ..." warn
     update
-    set bogus [catch {exec folders -fast -recurse} raw]
+    set bogus [catch {exec folders -all -fast -recurse} raw]
     set raw [split $raw \n]
     if {$bogus} {
 	set ix [lsearch -glob $raw "* * *"]
@@ -102,17 +160,20 @@ Please wait...
 
     set flist(allfolders) [FlistSort $raw]
     FlistSubFoldersInit
+    BgAction FlistUnseenFoldersInit FlistUnseenFoldersInit
     FlistCacheFolderList
     destroy .scanning
 }
 proc Flist_AddFolder { folder } {
     global flist
     if {[lsearch $flist(allfolders) $folder] >= 0} {
-	return
+	Exmh_Debug "Flist_AddFolder already has $folder"
+    } else {
+	lappend flist(allfolders) $folder
     }
-    lappend flist(allfolders) $folder
     set flist(allfolders) [FlistSort $flist(allfolders)]
     FlistSubFoldersInit
+    BgAction FlistUnseenFoldersInit FlistUnseenFoldersInit
     FlistCacheFolderList
     Fdisp_Redisplay
 }
@@ -124,6 +185,7 @@ proc Flist_DelFolder { folder } {
     }
     set flist(allfolders) [FlistSort [lreplace $flist(allfolders) $ix $ix]]
     FlistSubFoldersInit
+    BgAction FlistUnseenFoldersInit FlistUnseenFoldersInit
     FlistCacheFolderList
     Fdisp_Redisplay
 }
@@ -139,6 +201,24 @@ proc FlistCacheFolderList {} {
 	set flist(cacheFileMtime) [file mtime $flist(cacheFile)]
     }
 }
+proc FlistUnseenFoldersInit {} {
+    global flist mhProfile
+
+    set flist(unseenfolders) {}
+    foreach f $flist(allfolders) {
+        foreach pat $mhProfile(folder-unseen) {
+	    if {[string compare ! [string range $pat 0 0]] == 0} {
+		if [string match [string range $pat 1 end] $f] {
+			break
+		}
+	    }
+            if [string match $pat $f] {
+                lappend flist(unseenfolders) $f
+                break
+            }
+        }
+    }
+}
 proc FlistSubFoldersInit {} {
     global flist subFolders flistSubCache flistParents
 
@@ -174,6 +254,8 @@ proc Flist_FolderSet { {subfolder .} } {
     }
 }
 
+# The routines below here manage the unseen sequence state per folder.
+
 proc FlistSeq { folder sequence } {
     global mhProfile
     # Explode a sequence into a list of message numbers
@@ -204,70 +286,39 @@ proc FlistSeq { folder sequence } {
     }
     return $seq
 }
-proc Flist_UnseenMsgs { folder } {
-    global flist
-    set unseen [Mh_Unseen $folder]
-    set seen [Msg_Seen]
-    foreach id $seen {
-	set ix [lsearch $unseen $id]
-	if {$ix >= 0} {
-	    set unseen [lreplace $unseen $ix $ix]
-	}
-    }
-    if {[llength $unseen] == 0} {
-	catch {
-	    FlistUnseenFolder $folder
-	}
-    } else {
-	if ![info exists flist(new,$folder)] {
-	    Fdisp_HighlightUnseen $folder
-	}
-	set flist(newseq,$folder) $unseen
-	set flist(new,$folder) [llength $unseen]
-    }
-    return $unseen
-}
-proc Flist_NumUnseen { folder } {
-    global flist
-    if [info exists flist(new,$folder)] {
-	return $flist(new,$folder)
-    } else {
-	return 0
-    }
-}
-proc Flist_UnseenFolders {} {
-    global flist
-    return $flist(unseen)
-}
-proc FlistReset {} {
-    global flist
-    FlistInitVars
-}
+
+# Add unseen messages to the list for a given folder.
+# This has to be careful about already known unseen messages
+# and messages that have been read but not committed as read.
 proc Flist_AddUnseen {folder seq} {
     global flist exmh
-    set num [llength $seq]
-    if {$num <= 0} {
-	return
+    Exmh_Debug Flist_AddUnseen 1 $folder $seq
+
+    # Check overlap with already seen msgs and unseen messages already known
+    if ![info exists flist(newseq,$folder)] {
+	set flist(newseq,$folder) {}
+	set flist(new,$folder) 0
     }
-    if {$folder == $exmh(folder)} {
-	Flist_UpdateUnseen $folder $seq
-	return
+    if {[string compare $folder $exmh(folder)] == 0} {
+	set known [concat [Msg_Seen] $flist(newseq,$folder)]
+    } else {
+	set known $flist(newseq,$folder)
     }
-    incr flist(newMsgs) $num
-    set flist(new,$folder) $num
-    set flist(newseq,$folder) $seq
-    incr flist(newFolders)
-    lappend flist(unseen) $folder
-    lappend flist(unvisited) $folder
-    Fdisp_HighlightUnseen $folder
-}
-proc Flist_UpdateUnseen {folder seq} {
-    global flist exmh
-    # Check overlap with already seen msgs
-    foreach id [Msg_Seen] {
-	set ix [lsearch $seq $id]
-	if {$ix >= 0} {
-	    set seq [lreplace $seq $ix $ix]
+    # Subtract elements of $known from $seq
+    if {[llength $known] > [llength $seq]} {
+	set nseq {}
+	foreach id $seq {
+	    if {[lsearch $known $id] < 0} {
+		lappend nseq $id
+	    }
+	}
+	set seq $nseq
+    } else {
+	foreach id $known  {
+	    set ix [lsearch $seq $id]
+	    if {$ix >= 0} {
+		set seq [lreplace $seq $ix $ix]
+	    }
 	}
     }
     set num [llength $seq]
@@ -275,19 +326,23 @@ proc Flist_UpdateUnseen {folder seq} {
 	return
     }
     incr flist(newMsgs) $num
-    set flist(new,$folder) $num
-    set flist(newseq,$folder) $seq
+    incr flist(new,$folder) $num
+    set flist(newseq,$folder) [concat $flist(newseq,$folder) $seq]
+    Exmh_Debug Flist_AddUnseen 2 $folder $flist(newseq,$folder)
     if {[lsearch $flist(unseen) $folder] < 0} {
 	incr flist(newFolders)
 	lappend flist(unseen) $folder
     }
-    if {[lsearch $flist(unvisited) $folder] < 0} {
-	lappend flist(unvisited) $folder
+    if {[string compare $folder $exmh(folder)] != 0 &&
+	[lsearch $flist(unvisited) $folder] < 0} {
+	lappend flist(unvisitedNext) $folder
     }
     Fdisp_HighlightUnseen $folder
 }
 proc Flist_Done {} {
     global flist exmh
+
+    Exmh_Debug Flist_Done
     if {$flist(newMsgs) > 0} {
 	if {$flist(newMsgs) == 1} {set msg "msg"} else {set msg "msgs"}
 	if {$flist(newFolders) == 1} {set f "folder"} else {set f "folders"}
@@ -301,33 +356,82 @@ proc Flist_Done {} {
 	    Sound_Feedback $delta
 	}
     } else {
+	set flist(newMsgs) 0
 	Flag_NoUnseen
 	Exmh_Status "No unread msgs" blue
     }
     set flist(lastNewMsgs) $flist(newMsgs)
+    set flist(unvisited) [FlistSort $flist(unvisitedNext)]
+    set flist(active) 0
 }
 
 
+# Call Flist_UnseenUpdate from external sorting programs after
+# they add messages to a folder
+
+proc Flist_UnseenUpdate { folder } {
+    global exmh flist
+    Flist_UnseenMsgs $folder
+    if {[string compare $folder $exmh(folder)] == 0} {
+	Scan_FolderUpdate $folder
+    }
+    if {[lsearch $flist(unvisited) $folder] < 0} {
+	lappend flist(unvisited) $folder
+	set flist(unvisitedNext) $flist(unvisited)
+    }
+    # This wiggles the flag and sorts flist(unvisited)
+    Flist_Done
+}
+proc Flist_UnseenMsgs { folder } {
+    global flist
+    Flist_AddUnseen $folder [Mh_Unseen $folder]
+    return $flist(newseq,$folder)
+}
+proc Flist_NumUnseen { folder } {
+    global flist
+    if [info exists flist(new,$folder)] {
+	return $flist(new,$folder)
+    } else {
+	return 0
+    }
+}
+proc Flist_UnseenFolders {} {
+    global flist
+    return $flist(unseen)
+}
+
 # Flist enumerates folders that have unseen messages.
-proc Flist_FindUnseen {} {
-    Exmh_Debug Flist_FindUnseen end [time FlistFindUnseen]
+proc Flist_FindUnseen {{reset 0}} {
+    Exmh_Debug Flist_FindUnseen end [time [list FlistFindUnseen $reset]]
 }
 
-proc FlistFindUnseen {} {
+proc FlistFindStart {reset} {
+    global flist
+    if ![info exists flist(active)] {
+	set flist(active) 0
+    }
+    Exmh_Debug FlistFindStart active $flist(active)
+    if {$flist(active)} {
+	error "FlistActive"
+    }
+    set flist(active) 1
+    if {$reset} {
+	Fdisp_ClearHighlights
+	FlistResetVars
+    }
+    return 
+}
+proc FlistFindUnseen {reset} {
     global mhProfile flist curFolder
-    BgRPC FlistReset
+    Exmh_Debug FlistFindUnseen reset $reset
+    if [catch {BgRPC FlistFindStart $reset}] {
+	# Flist active
+	return
+    }
     FlistGetContext
-    foreach f $flist(allfolders) {
-	set check 0
-	foreach pat $mhProfile(folder-unseen) {
-	    if [string match $pat $f] {
-		set check 1
-		break
-	    }
-	}
-	if {! $check} {
-	    continue
-	}
+    set result {}
+    set keep {}
+    foreach f $flist(unseenfolders) {
 	set hit 0
 	foreach line $flist(context) {
 	    set line [split $line]
@@ -340,13 +444,21 @@ proc FlistFindUnseen {} {
 	    }
 	}
 	if {! $hit} {
-	    if {[catch {open $mhProfile(path)/$f/$mhProfile(mh-sequences)} in] == 0} {
-		set se \n[read $in]
-		if [regexp \n$mhProfile(unseen-sequence):\[^\n\]*\n $se line] {
-		    set seq [lindex [split $line :\n] 2]
-		    BgRPC Flist_AddUnseen $f [FlistSeq $f $seq]
+	    set path $mhProfile(path)/$f/$mhProfile(mh-sequences)
+	    if {![file exists $path] ||
+		([info exists flist(mtime,$f)] &&
+		 ([file mtime $path] <= $flist(mtime,$f)))} {
+		# No state to report
+	    } else {
+		if {[catch {open $path} in] == 0} {
+		    set se \n[read $in]
+		    if [regexp \n$mhProfile(unseen-sequence):\[^\n\]*\n $se line] {
+			set seq [lindex [split $line :\n] 2]
+			BgRPC Flist_AddUnseen $f [FlistSeq $f $seq]
+		    }
+		    close $in
+		    set flist(mtime,$f) [file mtime $path]
 		}
-		close $in
 	    }
 	}
     }
@@ -375,14 +487,18 @@ proc Flist_MsgSeen { msgid } {
 	    if {$flist(new,$exmh(folder)) == 0} {
 		FlistUnseenFolder $exmh(folder)
 	    }
+	    if {$flist(newMsgs) <  0} {
+		Exmh_Status "$flist(newMsgs) unseen!"
+		set flist(newMsgs) 0
+	    }
 	}
     }
 }
 proc FlistUnseenFolder { folder } {
     global flist
-    unset flist(new,$folder)
-    unset flist(newseq,$folder)
     Exmh_Debug FlistUnseenFolder $folder
+    catch {unset flist(new,$folder)}
+    catch {unset flist(newseq,$folder)}
     Fdisp_UnHighlightUnseen $folder
     set ix [lsearch $flist(unseen) $folder]
     if {$ix >= 0} {
@@ -396,55 +512,9 @@ proc FlistUnseenFolder { folder } {
     if {$ix >= 0} {
 	set flist(unvisited) [lreplace $flist(unvisited) $ix $ix]
     }
-}
-proc Flist_Update { folder {delete no} } {
-    global flist
-    if {$folder == {}} {
-	return
-    }
-    Exmh_Debug Flist_Update begin $folder
-    Exmh_Debug Flist_Update end [time [list FlistUpdate $folder $delete]]
-}
-proc FlistUpdate { folder {delete no} } {
-    global mhProfile flist
-    set dir $mhProfile(path)/$folder
-    set newseq {}
-    set hit 0
-    FlistGetContext
-    foreach line $flist(context) {
-	set line [split $line]
-	set key [lindex $line 0]
-	if {$key == "atr-$mhProfile(unseen-sequence)-$mhProfile(path)/$folder:"} {
-	    set seq [lindex [split $line :] 1]
-	    set newseq [FlistSeq $folder $seq]
-	    set hit 1
-	    break
-	}
-    }
-    if {! $hit} {
-	if {[catch {open $dir/$mhProfile(mh-sequences)} in] == 0} {
-	    while {[gets $in line] >= 0} {
-		if [regexp unseen: $line] {
-		    set seq [lindex [split $line :] 1]
-		    set newseq [FlistSeq $folder $seq]
-		    break
-		}
-	    }
-	    close $in
-	}
-    }
-    if [llength $newseq] {
-	Flist_UpdateUnseen $folder $newseq
-    } else {
-	Fdisp_UnHighlightUnseen $folder
-	set ix [lsearch $flist(unseen) $folder]
-	if {$ix >= 0} {
-	    set flist(unseen) [lreplace $flist(unseen) $ix $ix]
-	}
-	set ix [lsearch $flist(unvisited) $folder]
-	if {$ix >= 0} {
-	    set flist(unvisited) [lreplace $flist(unvisited) $ix $ix]
-	}
+    set ix [lsearch $flist(unvisitedNext) $folder]
+    if {$ix >= 0} {
+	set flist(unvisitedNext) [lreplace $flist(unvisitedNext) $ix $ix]
     }
 }
 
@@ -452,6 +522,17 @@ proc FlistSort { dirlist } {
     # Order the folder list according to a pattern template.
     # Patterns early in the list have higher priority.
 
+    # Hack to check against mh-e .folders file
+    if [regexp {\("\+} $dirlist] {
+	global flist
+	error \
+"Conflict with mh-e $flist(cacheFile).  Either remove it or override its name.
+The mh-e variable is mh-folder-list-filename.
+For exmh, set the variable flist(cacheFile) to another file.
+Add this to your user.tcl file (see exmh-custom man page for details).
+set flist(cacheFile) /usr/joe/Mail/.exmhfolders
+"
+    }
     global mhProfile
     set patterns $mhProfile(folder-order)
 
