@@ -12,12 +12,35 @@
 # any specification.
 
 proc Exmh {} {
-    global exmh
+    global exmh argv
 
     global tk_version
 
+    # Tk 4.0b3 bogosity
+    if [catch {tk colormodel .}] {
+	rename tk tk-orig
+	proc tk { option args } {
+	    switch -- $option {
+		colormodel {
+		    if {[winfo depth [lindex $args 0]] > 4} {
+			return color
+		    } else {
+			return monochrome
+		    }
+		}
+		default {
+		    return [eval {tk-orig $option} $args]
+		}
+	    }
+	}
+    }
+
     Mh_Init		;# Defines mhProfile
 
+    if {[lsearch $argv -mono] >= 0} {
+	# Do this early because it affects what X resources are used
+	tk colormodel . monochrome
+    }
     Preferences_Init ~/.exmh-defaults $exmh(library)/app-defaults
 
     # Add this preference to initialize exmh(userLibrary) and exmh(logEnabled)
@@ -32,6 +55,12 @@ after you add things to that directory."}
 "Debug information is recorded in a log that you can view
 from within exmh.  Turning off the log may save some
 memory usage.  You can enable the log temporarily."}
+	{exmh(logLines)	logLines 1000	{Max lines in debug log}
+"The log is implemented in a text widget.  This setting limits
+the number of lines kept in the log."}
+	{flist(debug)		flistDebug OFF	{Debug flist}
+"A listbox that displays the unseen and unvisited folder state
+is displayed to debug the flist module."}
     }
 
     ExmhArgv		;# snarf up command-line arguments
@@ -42,6 +71,7 @@ memory usage.  You can enable the log temporarily."}
 	auto_path_update $exmh(userLibrary)
     }
 
+    Mh_Preferences
     Sedit_BindInit	;# Text, Entry class bindings
     Widget_TextInit	;# Text scrolling
     ExmhLogInit		;# Enables debug loging
@@ -50,6 +80,8 @@ memory usage.  You can enable the log temporarily."}
 	puts stderr "User_Init: $err"
     }
 
+    catch {exec date} d
+    Audit "Startup $d $argv"
     # The order of the following mainly determines the way
     # their associated items appear in the Preferences dialog
     Sedit_Init		;# built in editor 
@@ -59,6 +91,7 @@ memory usage.  You can enable the log temporarily."}
     Ftoc_Init
     Msg_Init		;# Depends on Ftoc_Init, Buttons_Init
     Mime_Init
+    URI_Init
     Folder_Init		;# Sets exmh(folder)
     Inc_Init
     Exwin_Init
@@ -67,7 +100,11 @@ memory usage.  You can enable the log temporarily."}
     Fdisp_Init		;# After Flist and Fcache
     Sound_Init
     Faces_Init
+    Pgp_Init
+    Glimpse_Init
     Background_Init
+    fileselect_Init
+    Busy_Init
 
     wm protocol . WM_DELETE_WINDOW Exmh_Done
     Exwin_Layout
@@ -81,8 +118,8 @@ memory usage.  You can enable the log temporarily."}
 	wm iconify .
     }
     update
-    bind . <Unmap> ExmhUnmapped
-    bind . <Map> ExmhMapped
+    bind . <Unmap> {ExmhUnmapped %W}
+    bind . <Map> {ExmhMapped %W}
 
     Folder_Change $exmh(folder)
     busy ExmhJunk
@@ -94,11 +131,12 @@ proc ExmhJunk {} {
 }
 
 proc ExmhArgv {} {
-    global argc argv exmh
+    global argc argv exmh editor
     set extra {}
     set geo [option get . geometry Geometry]
     set icon [option get . iconposition IconPosition]
     set iconic [option get . iconic Iconic]
+    set editor(sedit!) 0	;# defeat accidental saving of override
     set bg_action {}
     for {set i 0} {$i < $argc} {incr i} {
 	set arg [lindex $argv $i]
@@ -126,6 +164,12 @@ proc ExmhArgv {} {
 		incr i
 		set exmh(bgPeriod) [lindex $argv $i]
 	    }
+	    "-sedit" {
+		set editor(sedit!) 1
+	    }
+	    "-*" {
+		catch {puts stderr "Unknown flag argument $arg"}
+	    }
 	    default {
 		lappend extra $arg
 	    }
@@ -138,13 +182,25 @@ proc ExmhArgv {} {
     }
     if {$geo != {}} {
 	if [catch {wm geometry . $geo} err] {
-	    puts stderr "-geometry $arg: $err"
+	    catch {puts stderr "-geometry $geo: $err"}
 	}
     }
-    if {$iconic != {}} {
-	set exmh(iconic) $iconic
-    } else {
-	set exmh(iconic) 0
+    switch $iconic {
+	""	{set exmh(iconic) 0}
+	True	-
+	TRUE	-
+	true	-
+	Yes	-
+	YES	-
+	yes	-
+	1	{set exmh(iconic) 1}
+	False	-
+	FALSE	-
+	false	-
+	no	-
+	NO	-
+	No	-
+	0	{set exmh(iconic) 0}
     }
     if {$icon != {}} {
 	Exwin_IconPosition . $icon
@@ -160,46 +216,48 @@ proc Exmh_Focus {} {
 proc ExmhResources {} {
     global exmh
     if {[tk colormodel .] == "color"} {
-	set exmh(c_st_normal) [option get . c_st_normal {}]
-	if {$exmh(c_st_normal) == {}} {set exmh(c_st_normal) blue}
-	set exmh(c_st_error) [option get . c_st_error {}]
-	if {$exmh(c_st_error) == {}} {set exmh(c_st_error) purple}
-	set exmh(c_st_warn) [option get . c_st_warn {}]
-	if {$exmh(c_st_warn) == {}} {set exmh(c_st_warn) red}
+	Preferences_Resource exmh(c_st_normal) c_st_normal blue
+	Preferences_Resource exmh(c_st_error) c_st_error purple
+	Preferences_Resource exmh(c_st_warn) c_st_warn red
+	Preferences_Resource exmh(c_st_background) c_st_background black
     } else {
-	set exmh(c_st_normal) [option get . c_st_normal {}]
-	if {$exmh(c_st_normal) == {}} {set exmh(c_st_normal) black}
+	Preferences_Resource exmh(c_st_normal) c_st_normal black
 	if {$exmh(c_st_normal) != "white" && $exmh(c_st_normal) != "black"} {
 	    set exmh(c_st_normal) black
 	}
 	set exmh(c_st_error) $exmh(c_st_normal)
 	set exmh(c_st_warn) $exmh(c_st_normal)
+	set exmh(c_st_background) $exmh(c_st_normal)
     }
 }
 
 proc Exmh_Status {string { level normal } } {
     global exmh exwin
-    if { $string == 0 } { set string $exmh(version) }
+    if {[string compare $string 0] == 0 } { set string $exmh(version) }
     if [info exists exwin(status)] {
-	case $level {
+	switch -- $level {
+	    warn	-
+	    error	-
+	    background	-
+	    normal	{ # do nothing }
 	    red		{set level warn}
 	    blue	{set level normal}
 	    purple	{set level error}
 	    "medium sea green" {set level background}
 	    default	{set level normal}
 	}
-	if ![info exists $exmh(c_st_$level)] {
+	if ![info exists exmh(c_st_$level)] {
 	    set exmh(c_st_$level) black
 	}
 	$exwin(status) configure -state normal
-	catch {$exwin(status) configure -fg $exmh(st_$level)}
+	catch {$exwin(status) configure -fg $exmh(c_st_$level)}
 	$exwin(status) delete 0 end
 	$exwin(status) insert 0 $string
 	$exwin(status) configure -state disabled
 	ExmhLog $string
 	update idletasks
     } else {
-	catch {puts stderr $string}
+	catch {puts stderr "exmh: $string"}
     }
 }
 proc Exmh_OldStatus {} {
@@ -212,22 +270,36 @@ proc Exmh_OldStatus {} {
 }
 
 proc Exmh_CheckPoint {} {
-    Scan_CacheUpdate
+    # This is really "folder change" CheckPoint
+    Exmh_Debug Scan_CacheUpdate [time Scan_CacheUpdate]
     Msg_CheckPoint
 }
 
 proc Exmh_Done {} {
-    global exmh
+    global exmh exwin
 
     if { [Ftoc_Changes "exit"] == 0} then {
-	Background_Wait
+	$exwin(mainButtons).quit config -state disabled
+	catch {exec date} d
+	Audit "Quit $d"
 	Exmh_Status "Checkpointing state" red
-	Exmh_CheckPoint
-	Fcache_CheckPoint
-	Background_Cleanup
-	Mime_Cleanup
-	Exwin_CheckPoint
-	Sedit_CheckPoint
+	if [info exists exmh(newuser)] {
+	    PreferencesSave nodismiss	;# Save tuned parameters
+	}
+	Background_Wait
+	foreach cmd {Background_Cleanup
+		    Hook_CheckPoint Sedit_CheckPoint Aliases_CheckPoint
+		    Exmh_CheckPoint Fcache_CheckPoint
+		    Audit_CheckPoint Mime_Cleanup
+		    Exwin_CheckPoint Pgp_CheckPoint
+		    } {
+	    if {[info command $cmd] != {}} {
+		Exmh_Status $cmd
+		if [catch $cmd err] {
+		    catch {puts stderr "$cmd: $err"}
+		}
+	    }
+	}
 	destroy .
     }
 }
@@ -236,12 +308,16 @@ proc Exmh_Abort {} {
     destroy .
 }
 
-proc ExmhUnmapped {} {
+proc ExmhUnmapped {w} {
     # This triggers auto-commit
-    Ftoc_Changes iconified
+    if {$w == "."} {
+	Ftoc_Changes iconified
+    }
 }
-proc ExmhMapped {} {
-    Inc_Mapped
+proc ExmhMapped {w} {
+    if {$w == "."} {
+	Inc_Mapped
+    }
 }
 
 #### Exmh_Debugging
@@ -264,7 +340,7 @@ proc ExmhLogInit {} {
     set exmh(logWindow) 0
 }
 proc ExmhLog { stuff } {
-    global exmh
+    global exmh tk_version
     if {![info exists exmh(logInit)]} {
 	return
     }
@@ -287,17 +363,42 @@ proc ExmhLog { stuff } {
 	Widget_AddBut $exmh(logTop).but save "Save To File" ExmhLogSave
 	set exmh(logYview) 1
 	Widget_CheckBut $exmh(logTop).but yview "View Tail" exmh(logYview)
-	set exmh(log) [Widget_Text $exmh(logTop) 20 -setgrid true]
+	set exmh(log) [Widget_Text $exmh(logTop) 20 \
+		-setgrid true -yscroll {.log.sv set} ]
 	#
-	$exmh(log) configure -yscroll {.log.sv set}
+	# Set up Tcl command type-in
+	#
+	Widget_BindEntryCmd $exmh(log) <Control-c>  \
+	    "focus $exmh(logTop).cmd.entry"
+	Widget_Bindtags $exmh(log) [list $exmh(log) Text $exmh(logTop) all]
+	Widget_BeginEntries 4 80 Exmh_DoCommand
+	Widget_LabeledEntry $exmh(logTop).cmd Tcl: exmh(command)
     }
     if [info exists exmh(log)] {
 	catch {
+#	    $exmh(log) insert end " [bw_delta] "
 	    $exmh(log) insert end $stuff
 	    $exmh(log) insert end \n
 	    if {$exmh(logYview)} {
 		$exmh(log) yview -pickplace "end - 1 lines"
 	    }
+	    scan [$exmh(log) index end] %d numlines
+	    if {$numlines > $exmh(logLines)} {
+		set numlines [expr {$numlines - $exmh(logLines)}]
+		$exmh(log) delete 1.0 $numlines.0
+	    }
+	}
+    }
+}
+proc LogStart { what } {
+    global exmh
+    if [info exists exmh(log)] {
+	catch {
+	    $exmh(log) insert end "\n[bw_delta $what]"
+	    $exmh(log) insert end \n
+	    if {$exmh(logYview)} {
+		$exmh(log) yview -pickplace "end - 1 lines"
+	    }
 	}
     }
 }
@@ -316,7 +417,7 @@ proc ExmhLogTrunc {} {
 proc ExmhLogSave {} {
     global exmh
     for {set id 0} {$id < 100} {incr id} {
-	set name /tmp/exmhlog.$id
+	set name [Env_Tmp]/exmhlog.$id
 	if ![file exists $name] {
 	    if ![catch {open $name w} logfile] {
 		break
@@ -326,7 +427,7 @@ proc ExmhLogSave {} {
     if [catch {
 	puts $logfile [$exmh(log) get 1.0 end]
 	close $logfile
-	Exmh_Status "Saved log in /tmp/exmhlog.$id" blue
+	Exmh_Status "Saved log in [Env_Tmp]/exmhlog.$id" blue
     } msg] {
 	Exmh_Status "Cannot save log: $msg" purple
     }
@@ -347,6 +448,24 @@ proc TraceInfo {} {
 proc DoNothing { args } {
     return ""
 }
-
-
-
+proc Exmh_DoCommand {} {
+    global exmh
+    if {[string length $exmh(command)] == 0} {
+	return
+    }
+    set t $exmh(log)
+    $t insert end $exmh(command)\n
+    update idletasks
+    if [catch {uplevel #0 $exmh(command)} result] {
+	global errorInfo
+	$t insert end "ERROR\n$errorInfo\n\n"
+    } else {
+	$t insert end $result\n\n
+    }
+    global tk_version
+    if {$tk_version >= 4.0} {
+	$t see end
+    } else {
+	$t yview -pickplace end
+    }
+}
