@@ -23,14 +23,22 @@ proc Msg_Reset { numMsgs {folder {}} } {
     set msg(seenOld) {}			;# seen, then deleted or moved
     set msg(dpy)  {}			;# Currently displayed message
     set msg(id) [Mh_Cur $folder]	;# pick current message
+    set msg(path) ""			;# File pathname of current message
     Buttons_Current [expr {$msg(id) != {}}]	;# Enable/disable buttons
     Ftoc_Reset $numMsgs $msg(id) $folder	;# Reset display
 }
 proc Msg_CheckPoint {} {
     # push current MH state to disk
     global exmh msg
-    if {$msg(id) != {}} {
-	Mh_SetCur $exmh(folder) $msg(id)
+    if {$msg(id) != ""} {
+	set cur $msg(id)
+    } else {
+	set cur ""
+    }
+    if {$cur != ""} {
+	Mh_SetCur $exmh(folder) $cur
+    } else {
+	Mh_ClearCur $exmh(folder)
     }
     if {$msg(seen) != {}} {
 	Mh_MarkSeen $exmh(folder) $msg(seen)
@@ -54,23 +62,25 @@ proc Msg_Pick { line {show show} } {
 proc Msg_ShowCurrent { {show show} } {
     global msg
     if {$msg(id) != {}} {
+	set msg(dpy) {}	;# force redisplay
 	Msg_Change $msg(id) $show
 	return 1
     } else {
 	Msg_ClearCurrent
+	Ftoc_Yview end
 	return 0
     }
 }
 proc Msg_ShowUnseen { {show show} } {
     global exmh
-    set unseen [Flist_UnseenMsgs $exmh(folder)]
-    if {[llength $unseen] > 0} then {
-	Msg_Change [lindex $unseen 0] show
-	return 1
-    } else {
-	Msg_ClearCurrent
-	return 0
+    foreach id [Flist_UnseenMsgs $exmh(folder)] {
+	if {![Ftoc_Marked $id]} {
+	    Msg_Change $id show
+	    return 1
+	}
     }
+    Msg_ClearCurrent
+    return 0
 }
 proc Msg_ClearCurrent { } {
     global msg exmh
@@ -79,10 +89,12 @@ proc Msg_ClearCurrent { } {
     Mh_ClearCur $exmh(folder)
     MsgClear
     Buttons_Current 0
+    Uri_ClearCurrent
 }
 proc MsgClear {} {
-    global exwin
+    global exwin msg
     Label_Message ""
+    set msg(dpy) {}
     $exwin(mtext) configure -state normal
     $exwin(mtext) delete 0.0 end
     $exwin(mtext) configure -state disabled
@@ -92,7 +104,7 @@ proc Msg_ShowSomething {} {
     global exmh msg
     set order {unseen cur}
     foreach pick $order {
-	if {[catch {exec pick +$exmh(folder) $pick} tmp] == 0} then {
+	if {[catch {MhExec pick +$exmh(folder) $pick} tmp] == 0} then {
 	    Msg_Change [lindex $tmp 0] show
 	    return
 	}
@@ -110,13 +122,20 @@ proc Msg_ShowWhat { {what last} {show show} } {
 	return 1
     }
 }
+proc Msg_First { {show noshow} } {
+    Msg_Change {} $show first
+}
+
+proc Msg_Last { {show noshow} } {
+    Msg_Change {} $show last
+}
 
 proc Msg_Change {msgid {show show} {line {}} } {
     Exmh_Debug Msg_Change id=$msgid line=$line
     Exmh_Debug Msg_Change [time [list MsgChange $msgid $show $line]]
 }
 proc MsgChange {msgid {show show} {line {}} } {
-    global exmh exwin msg
+    global exmh exwin msg mhProfile
 
     if {$msgid != {}} {
 	# Allow null msgid from Msg_ShowWhat, which supplies line instead
@@ -132,6 +151,7 @@ proc MsgChange {msgid {show show} {line {}} } {
 	    Buttons_Current 1
 	}
 	set msg(id) $msgid
+	set msg(path) $mhProfile(path)/$exmh(folder)/$msg(id)
 	if {$show == "show"} {
 	    MsgShow $msgid
 	} else {
@@ -173,33 +193,88 @@ proc Msg_Seen {} {
 
 
 # Message operations.
-# The ones that generate messages take a callback that sets up
-# the draft folder so you can override this initial setup.
+# These take two forms of arguments.  The original form is a single
+# argument that is the name of a hook procedure.  The new form is
+# a set of arguments for the underlying MH command.  The distinction is
+# made by seeing if the argument is the name of a Tcl command, if
+# not it is assumed to be arguments to the MH command.
 
-proc Msg_Compose { {compProc Mh_CompSetup} } {
-    if [catch {$compProc} err] {			;# Setup draft msg
-	Exmh_Status "${compProc}: $err" purple
+proc Msg_Compose { args } {
+    global exmh msg	;# allow args to include $exmh(folder) $msg(id)
+    if {[string length $args] == 0} {
+	set args Mh_CompSetup
+    }
+    if {[string compare [info command $args] $args] == 0} {
+	# Old interface with hook procedure
+	if [catch {$args} err] {			;# Setup draft msg
+	    Exmh_Status "$args: $err" purple
+	    return
+	}
     } else {
-	Edit_Draft					;# Run the editor
+	if [catch {
+	    Exmh_Status "comp $args"
+	    eval {MhExec comp -nowhatnowproc} $args
+	} err] {
+	    Exmh_Status "comp: $err"
+	    return
+	}
+    }
+    Edit_Draft					;# Run the editor
+}
+proc Msg_CompUse {folder id} {
+    global mhProfile
+    if {[string compare $folder $mhProfile(draft-folder)] == 0} {
+	Mh_SetCur $mhProfile(draft-folder) $id
+	Msg_Compose $id -use
+    } else {
+	Msg_Compose +$folder $id
     }
 }
-
-proc Msg_Reply { {replyProc Mh_ReplySetup} } {
+proc Msg_Reply { args } {
     global exmh msg
+    if {[string length $args] == 0} {
+	set args Mh_ReplySetup
+    }
 
     if [MsgOk $msg(id) m] {
 	Mh_AtLink $exmh(folder) $m			;# Symlink to @
-	if [catch {$replyProc $exmh(folder) $m} err] {	;# Setup draft msg
-	    Exmh_Status "${replyProc}: $err" purple
-	    Mh_AtLinkCleanup				;# Nuke @ link
+	set edit 1
+	if {[string compare [info command $args] $args] == 0} {
+	    # Old interface with hook procedure
+	    if [catch {$args $exmh(folder) $m} err] {	;# Setup draft msg
+		Exmh_Status "${args}: $err" purple
+		Mh_AtLinkCleanup				;# Nuke @ link
+		return
+	    }
 	} else {
-	    Edit_Draft					;# Run the editor
+	    Exmh_Status "repl $args" purple
+	    if [catch {
+		set ix [lsearch $args -noedit]
+		if {$ix >= 0} {
+		    set edit 0
+		    set args [lreplace $args $ix $ix]
+		}
+		eval {MhExec repl +$exmh(folder) $m -nowhatnowproc} $args
+		MhAnnoSetup $exmh(folder) $m repl
+	    } err] {	;# Setup draft msg
+		Exmh_Status "repl: $err" purple
+		Mh_AtLinkCleanup				;# Nuke @ link
+		return
+	    }
 	}
-    }
+	if {$edit} {
+	    Edit_Draft					;# Run the editor
+	} else {
+	    Edit_Done send				;# Just send it
+	}
+   }
 }
 
-proc Msg_Forward { {forwProc Mh_ForwSetup} } {
+proc Msg_Forward { args } {
     global exmh msg
+    if {[string length $args] == 0} {
+	set args Mh_ForwSetup
+    }
 
     set ids {}
     Ftoc_Iterate line {
@@ -209,25 +284,65 @@ proc Msg_Forward { {forwProc Mh_ForwSetup} } {
 	}
     }
     if {[llength $ids] > 0} {
-	if [catch {$forwProc $exmh(folder) $ids} err] {	;# Setup draft msg
-	    Exmh_Status "${forwProc}: $err" purple
-	    Mh_AtLinkCleanup				;# Nuke @ link
-	} else {
-	    Edit_Draft					;# Run the editor
+	global mhProfile
+	set mime 0
+	if [info exists mhProfile(forw)] {
+	    if {[lsearch $mhProfile(forw) -mime] >= 0} {
+		set mime 1
+	    }
 	}
-    }
+	if {[string compare [info command $args] $args] == 0} {
+	    # Old interface with hook procedure
+	    if [catch {$args $exmh(folder) $ids} err] {	;# Setup draft msg
+		Exmh_Status "${args}: $err" purple
+		return
+	    }
+	}  else {
+	    Exmh_Status "forw +$exmh(folder) $ids $args"
+	    if [catch {
+		if {[lsearch $args -mime] >= 0} {
+		    set mime 1
+		}
+		eval {MhExec forw +$exmh(folder)} $ids -nowhatnowproc $args
+		MhAnnoSetup $exmh(folder) $ids forw
+	    } err] {
+		Exmh_Status "forw: $err" purple
+		return
+	    }
+	}
+	# sedit hack
+	global sedit
+	set old $sedit(mhnDefault)
+	if {$mime} {set sedit(mhnDefault) 1}
+	Edit_Draft					;# Run the editor
+	set sedit(mhnDefault) $old
+   }
 }
 
-proc Msg_Dist { {distProc Mh_DistSetup} } {
+proc Msg_Dist { args } {
     global exmh msg
+    if {[string length $args] == 0} {
+	set args Mh_DistSetup
+    }
 
     if [MsgOk $msg(id) m] {
-      if [catch {$distProc $exmh(folder) $m} err] {   ;# Setup draft msg
-          Exmh_Status "${distProc}: $err" purple
-          Mh_AtLinkCleanup                            ;# Nuke @ link
-      } else {
-          Edit_Draft                                  ;# Run the editor
-      }
+	if {[string compare [info command $args] $args] == 0} {
+	    # Old interface with hook procedure
+	    if [catch {$args $exmh(folder) $m} err] {   ;# Setup draft msg
+		Exmh_Status "${args}: $err" purple
+		return
+	    }
+	}  else {
+	    if [catch {
+		Exmh_Status "dist +$exmh(folder) $m"
+		eval {MhExec dist +$exmh(folder) $m} -nowhatnowproc $args
+		MhAnnoSetup $exmh(folder) $m dist
+	    } err] {
+		Exmh_Status "dist: $err" purple
+		return
+	    }
+	}
+	Edit_Draft                                  ;# Run the editor
     }
 }
  
@@ -242,21 +357,20 @@ proc MsgOk { number msgvar } {
     }
 }
 
-
-
-proc Msg_Remove { {rmProc Ftoc_RemoveMark} } {
+proc Msg_Remove { {rmProc Ftoc_RemoveMark} {show show} } {
     Exmh_Debug Msg_Remove $rmProc
     Ftoc_Iterate line {
 	set msgid [Ftoc_MsgNumber $line]
 	Exmh_Debug Msg_Remove l=$line m=$msgid
 	$rmProc $line $msgid
-	# to avoid MH mark bug
-#	MsgUnSeen $msgid
     }
     if {[Ftoc_PickSize] == 1} {
-	Ftoc_Next show soft
+	Ftoc_NextImplied $show
     }
 }
+proc Msg_RemoveNoshow { {rmProc Ftoc_RemoveMark} } {
+    Msg_Remove $rmProc noshow
+}
 proc Msg_RemoveById { msgid {rmProc Ftoc_Delete} } {
     global msg
     set line [Ftoc_FindMsg $msgid]
@@ -266,7 +380,7 @@ proc Msg_RemoveById { msgid {rmProc Ftoc_Delete} } {
 	Msg_ClearCurrent
     }
 }
-proc Msg_Move { {moveProc Ftoc_MoveMark} } {
+proc Msg_Move { {moveProc Ftoc_MoveMark} {advance 1} {show show} } {
     global exmh
 
     if {$exmh(target) == ""} {
@@ -274,26 +388,29 @@ proc Msg_Move { {moveProc Ftoc_MoveMark} } {
 	return
     }
     if { $exmh(target) != $exmh(folder)} then {
-	Exmh_Debug Msg_Move picksize=[Ftoc_PickSize]
 	Ftoc_Iterate line {
 	    set msgid [Ftoc_MsgNumber $line]
-	    Exmh_Debug $moveProc l=$line id=$msgid
 	    $moveProc $line $msgid
-	    # To avoid MH mark bug
-#	    MsgUnSeen $msgid
 	}
-	if {[Ftoc_PickSize] == 1} {
-	    Ftoc_Next show soft
+	Exmh_Status "=> $exmh(target)"
+	if {[Ftoc_Advance $advance] && ([Ftoc_PickSize] == 1)} {
+	    Ftoc_NextImplied $show
 	}
     } else {
 	Exmh_Status "Move requires target folder != current folder"
     }
 }
-
-proc Msg_Clip {} {
+proc Msg_MoveNoshow { {moveProc Ftoc_MoveMark} } {
+    Msg_Move $moveProc 1 noshow
+}
+proc Msg_Clip { {folder {}}  {id {}} } {
     # "Tear off" a message into a top-level text widget
     global mhProfile exmh msg exwin
-    if {$msg(id) == {}} {
+
+    if {$folder == {}} {set folder $exmh(folder)}
+    if {$id     == {}} {set id     $msg(id)}
+
+    if {$id == {}} {
 	Exmh_Status "Select a message to clip first" red
 	return
     }
@@ -302,14 +419,21 @@ proc Msg_Clip {} {
     } else {
 	incr msg(tearid)
     }
-    set self [Widget_Toplevel .tear$msg(tearid) "$exmh(folder) $msg(id)" Clip]
+    set self [Widget_Toplevel .tear$msg(tearid) "$folder $id" Clip]
 
     Widget_Frame $self but Menubar {top fill}
     Widget_AddBut $self.but quit "Dismiss" [list destroy $self]
-    Widget_Label $self.but label {left fill} -text $exmh(folder)/$msg(id)
+    Widget_Label $self.but label {left fill} -text $folder/$id
     set t [Widget_Text $self $exwin(mtextLines) -cursor xterm -setgrid true]
     Msg_Setup $t
-    MsgShowInText $t $mhProfile(path)/$exmh(folder)/$msg(id)
+    if [MsgShowInText $t $mhProfile(path)/$folder/$id] {
+        foreach cmd [info commands Hook_MsgClip*] {
+            if [catch {$cmd $mhProfile(path)/$folder/$id $t} err] {
+                SeditMsg $t "$cmd $err"
+            }
+        }
+    }
+
 }
 proc Msg_FindMatch {L string} {
     global exwin
@@ -322,11 +446,15 @@ proc Msg_BurstDigest {} {
 	Exmh_Status "No message selected to burst" purple
 	return
     }
+    if {[Ftoc_Changes "Burst Digest"] != 0} {
+	# Pending changes and no autoCommit
+	return
+    }
+
     Exmh_Status "Bursting message $msg(id) in $exmh(folder)..." blue
 
     # burst the digest; catch the output
-    Msg_CheckPoint ;# clear cached seen/unseen state
-    if [catch { exec burst -verbose $msg(id) +$exmh(folder)} out] {
+    if [catch { MhExec burst -verbose $msg(id) +$exmh(folder)} out] {
 	Exmh_Status "Error bursting digest: $out"
     } else {
 	# burst OK, split up the output
@@ -340,12 +468,17 @@ proc Msg_BurstDigest {} {
 	# mark new messages as unread
 	Exmh_Debug burst created msgs $allids
 	if {$allids != {}} {
-	    eval { exec mark +$exmh(folder) -sequence $mhProfile(unseen-sequence) } $allids
+	    eval { MhExec mark +$exmh(folder) -sequence $mhProfile(unseen-sequence) } $allids
 	}
-	# rescan to pick them up
-	Scan_Folder $exmh(folder)
-	Msg_ClearCurrent
+	# rescan to pick them up, make sure Commit is done.
+	Background_Wait
 	Exmh_Status "Bursting message $msg(id) in $exmh(folder)...done" blue
+	Scan_FolderUpdate $exmh(folder)
+	if {$allids != {}} {
+	    Msg_Change [lindex $allids 0]
+	} else {
+	    Msg_ClearCurrent
+	}
     }
 }
 proc Msg_Save {} {
@@ -374,4 +507,78 @@ proc Msg_Save {} {
     }
 }
 
+proc Msg_Edit {} {
+    global exmh msg editor
+    Exmh_Status "Editing $exmh(folder)/$msg(id)"
+    #
+    # Hack because exmh-async isn't appropriate in this case.
+    #
+    if {$editor(sedit!)} {
+	set edittype sedit
+    } else {
+	set edittype prog
+    }
+    if [regsub {^([ 	]*)exmh-async(.*)$} $editor($edittype) {\2} newprog] {
+	set cmd [split [join [string trimright $newprog "& \t"]]]
+	Exmh_Status "Starting $cmd ..." warn
+	if [catch {eval exec $cmd $msg(path) &} err] {
+	    Exmh_Status $err error
+	}
+    } else {
+	EditStart $msg(path) $edittype
+    }
+}
 
+proc Msg_UUdecode {} {
+    global exmh msg mhProfile
+    set name [FSBox "Select file to decode into:" ]
+    if {$name != {}} {
+	Mime_Uudecode $msg(path) $name
+    } else {
+	Exmh_Status "uudecode canceled"
+    }
+}
+
+proc Msg_MarkUnseen {} {
+    global exmh
+    Msg_CheckPoint
+    Ftoc_Iterate line {
+	set msgid [Ftoc_MsgNumber $line]
+	Mh_MarkUnseen $exmh(folder) $msgid
+    }
+    Msg_ClearCurrent
+    Ftoc_ClearCurrent
+    Ftoc_ShowUnseen $exmh(folder)
+}
+
+proc Msg_ReplyHelp {} {
+    Help Reply "Defining Reply Buttons and Menu Entries"
+}
+
+proc Msg_PageOrNext {} {
+    global exwin
+    Widget_TextPageOrNext $exwin(mtext)
+}
+proc Msg_PageDown {} {
+    global exwin
+    Widget_TextPageDown $exwin(mtext)
+}
+proc Msg_PageUp {} {
+    global exwin
+    Widget_TextPageUp $exwin(mtext)
+}
+proc Msg_LineDown {} {
+    global exwin
+    Widget_TextLineDown $exwin(mtext)
+}
+proc Msg_LineUp {} {
+    global exwin
+    Widget_TextLineUp $exwin(mtext)
+}
+proc Msg_CopySelection {} {
+    global exwin sedit
+    catch {set sedit(killbuf) [$exwin(mtext) get sel.first sel.last]}
+}
+proc Msg_Trash { {trashFolder TRASH} } {
+    Folder_TargetMove $trashFolder
+}
