@@ -38,16 +38,22 @@ proc Preferences_Init { userDefaults appDefaults } {
     set pref(panes) {}
     set pref(userDefaults) $userDefaults
     set pref(appDefaults) $appDefaults
+    set pref(localDefaults) \
+	"[file dirname $appDefaults]/local.[file tail $appDefaults]"
 
-    PreferencesReadFile $appDefaults startup
-    if [file exists $userDefaults] {
-	PreferencesReadFile $userDefaults user
-    }
+    PreferencesReadFile $pref(appDefaults) startup
+    PreferencesReadFile $pref(localDefaults) 50
+    PreferencesReadFile $pref(userDefaults) user
+
+    Preferences_Resource pref(helpInOneWindow) helpInOneWindow 1
 }
 proc PreferencesReadFile { basename level } {
-    if [catch {option readfile $basename $level} err] {
-	Exmh_Status "Error in $basename: $err"
+    if [file exists $basename] {
+	if [catch {option readfile $basename $level} err] {
+	    Exmh_Status "Error reading '$basename': $err"
+	}
     }
+
     if {[tk colormodel .] == "color"} {
 	if [file exists $basename-color] {
 	    if [catch {option readfile $basename-color $level} err] {
@@ -72,7 +78,9 @@ proc Preferences_Add { id text prefs } {
 	set pref($id,prefs) $prefs
 	set pref($id,text) $text
     } else {
-	lappend pref($id,prefs) $prefs
+	foreach p $prefs {
+	    lappend pref($id,prefs) $p
+	}
 	append pref($id,text) \n$text
     }
 
@@ -87,33 +95,24 @@ proc Preferences_Add { id text prefs } {
 	Exmh_Debug Pref_Add $varName $value
 	if {$value == {}} {
 	    # Set variables that are still not set
-	    if {[llength $default] > 1} {
-		if {[lindex $default 0] == "CHOICE"} {
-		    PrefValueSet $varName [lindex $default 1]
-		} else {
-		    PrefValueSet $varName $default
-		}
-	    } else {
-		if {$default == "OFF" || $default == "ON"} {
-		    # This is a boolean
-		    if {$default == "OFF"} {
-			PrefValueSet $varName 0
-		    } else {
-			PrefValueSet $varName 1
-		    }
-		} else {
-		    # This is a string or numeric
-		    PrefValueSet $varName $default
-		}
+	    switch -regexp -- $default {
+		^ON$		{PrefValueSet $varName 1}
+		^OFF$		{PrefValueSet $varName 0}
+		"^CHOICE "	{PrefValueSet $varName [lindex $default 1]}
+		default		{PrefValueSet $varName $default}
 	    }
 	} else {
 	    # Warp booleans to 0 or 1
 	    if {$default == "OFF" || $default == "ON"} {
-		Exmh_Debug Pref Check $varName $value
-		case $value {
-		    {0 1} { # ok as is }
-		    {true True TRUE} { PrefValueSet $varName 1}
-		    {false False FALSE} {PrefValueSet $varName 0}
+		switch -- $value {
+		    0 -
+		    1 		{ # ok as is }
+		    true -
+		    True -
+		    TRUE 	{PrefValueSet $varName 1}
+		    false -
+		    False -
+		    FALSE 	{PrefValueSet $varName 0}
 		    default {
 			catch {puts stderr "Bogus boolean value $value for Xresource $xresName"}
 			PrefValueSet $varName 0
@@ -167,47 +166,7 @@ proc PreferencesDelete {} {
 }
 
 proc PreferencesHelp {} {
-    if [Exwin_Toplevel .prefhelp "Preferences Help" Help] {
-	wm group .prefhelp .pref
-	Widget_Label .prefhelp.but label {left fill} \
-	    -text "Help for Preferences"
-	set helptext "
-The Preference settings are divided into a number of
-subsections.  These correspond to different modules of
-the Exmh implementation.  Use the buttons in the main
-Preferences window to select various sections.
-
-Information about each setting will appear if you click on
-the short description of the preference item.  You can make
-the info window go away by clicking on another label or
-by clicking inside the info window.
-
-There are three types of options you can set through the Preferences dialog.
-Choices are represented by radio-style buttons where only one button
-in the set can be enabled at once.  These take effect immediately.
-Booleans are represented by check-sytle buttons.  If the checkbox is
-dark, then the option is turned on.  This takes effect immediately.
-
-Numeric and filename settings have entry widgets in which you can
-type in a new value.  The new value takes effect when you type
-<Return>, or when you click \"Save\".
-
-The Save button will save your settings in a .exmh-defaults file
-in your home directory.  This is an Xresources-style file.
-
-The Reset button will restore the settings from defaults and your
-last saved pref.
-
-The Dismiss button leaves the settings as they are displayed
-and removes the pref window.  Thus, those settings
-that \"take effect immediately\" are still in effect but they
-are not saved to your pref file.
-"
-	set numLines [llength [split $helptext \n]]
-	if {$numLines > 25} {set numLines 25}
-	set t [Widget_Text .prefhelp $numLines -setgrid true]
-	$t insert 1.0 $helptext
-    }
+    Help Preferences "Help for Preferences"
 }
 
 proc Preferences_Dialog {} {
@@ -255,6 +214,9 @@ proc PreferencesSectionDialog { id } {
 	wm protocol .pref$ix WM_DELETE_WINDOW [list PreferencesDismiss $ix]
 	wm minsize .pref$ix 25 2
 	Widget_AddBut $buttons reset Reset [list PreferencesReset $id]
+	if $pref(helpInOneWindow) {
+	    Widget_AddBut $buttons help Help [list PreferencesPaneHelp $id]
+	}
 	Widget_AddBut $buttons next Next [list PreferencesNext $ix] {left}
 	Widget_AddBut $buttons prev Prev [list PreferencesNext $ix -1] {left}
 
@@ -266,10 +228,11 @@ proc PreferencesSectionDialog { id } {
 	$body configure -borderwidth 10
 	set body [Widget_Frame $body body Body]
 
-	set txt [Widget_Text [Widget_Frame $body text] 4]
-	$txt insert 1.0 $pref($id,text)
-	$txt configure -state disabled
-#	Widget_Message $body message -text $pref($id,text)
+	if !$pref(helpInOneWindow) {
+	    set txt [Widget_Text [Widget_Frame $body text] 4]
+	    $txt insert 1.0 $pref($id,text)
+	    $txt configure -state disabled
+	}
 	set maxWidth 0
 	foreach item $pref($id,prefs) {
 	    set len [string length [PrefComment $item]]
@@ -278,7 +241,7 @@ proc PreferencesSectionDialog { id } {
 	    }
 	}
 	foreach item $pref($id,prefs) {
-	    PreferencesDialogItem $body $item $maxWidth
+	    PreferencesDialogItem $body $id $item $maxWidth
 	}
     }
     Exwin_ToplevelFocus .pref$ix none
@@ -298,29 +261,36 @@ proc PreferencesNext { ix {i 1}} {
     }
 }
 
-proc PreferencesDialogItem { frame item width } {
+proc PreferencesDialogItem { frame id item width } {
     global pref
     incr pref(uid)
     set f [Widget_Frame $frame p$pref(uid) Preference {top fill}]
     Widget_Label $f label {left fill} \
 	-text [PrefComment $item] -width $width
-    bind $f.label <1> [list PreferencesItemHelp  %X %Y [PrefHelp $item]]
+    
+    if $pref(helpInOneWindow) {
+	bind $f.label <1> [list PreferencesPaneHelp $id [PrefXres $item]]
+    } else {
+	bind $f.label <1> [list PreferencesItemHelp  %X %Y [PrefHelp $item]]
+    }
 
     set default [PrefDefault $item]
-    if {([llength $default] > 1) && ([lindex $default 0] == "CHOICE")} {
-	# >1 This is a list of choices
-	foreach choice [lreplace $default 0 0] {
-	    incr pref(uid)
-	    Widget_RadioBut $f c$pref(uid) $choice [PrefVar $item] {left}
-	}
-    } else {
-	if {$default == "OFF" || $default == "ON"} {
+    switch -regexp -- $default {
+	^(ON|OFF)$	{
 	    # This is a boolean
 	    set varName [PrefVar $item]
 	    Widget_CheckBut $f check "On" $varName {left}
 	    $f.check config -command [list PrefBooleanFixup $f.check $varName]
 	    PrefBooleanFixup $f.check $varName
-	} else {
+	}
+	"^CHOICE "	{
+	    # This is a list of choices
+	    foreach choice [lreplace $default 0 0] {
+		incr pref(uid)
+		Widget_RadioBut $f c$pref(uid) $choice [PrefVar $item] {left}
+	    }
+	}
+	default	 {
 	    # This is a string or numeric
 	    global PrefEntry
 	    Widget_Entry $f entry {left fill expand} -width 10
@@ -328,7 +298,8 @@ proc PreferencesDialogItem { frame item width } {
 
 	    set varName [PrefVar $item]
 	    $f.entry insert 0 [uplevel #0 [list set $varName]]
-	    bind $f.entry <Return> [list PrefEntrySet %W $varName]
+	    Widget_BindEntryCmd $f.entry <Return> \
+		[list PrefEntrySet %W $varName]
 	}
     }
 }
@@ -359,7 +330,7 @@ proc PreferencesNukeItemHelp { t } {
     destroy $t
 }
 
-proc PreferencesSave {} {
+proc PreferencesSave { {nodismiss {}} } {
     global pref PrefEntry
     set newstuff {}
     foreach id $pref(panes) {
@@ -375,7 +346,9 @@ proc PreferencesSave {} {
     }
     Preferences_RewriteSection "Lines below here automatically added" "End Preferences State" $newstuff
     PreferencesReset
-    PreferencesDismiss
+    if {$nodismiss == {}} {
+	PreferencesDismiss
+    }
     Background_Preferences
 }
 proc Preferences_RewriteSection { boundary1 boundary2 newstuff } {
@@ -470,6 +443,7 @@ proc PreferencesReset { {id_in {}} } {
     # Re-read user defaults
     option clear
     PreferencesReadFile $pref(appDefaults) startup
+    PreferencesReadFile $pref(localDefaults) 50
     PreferencesReadFile $pref(userDefaults) user
     # Now set variables
     if {$id_in == {}} {
@@ -487,28 +461,77 @@ proc PreferencesReset { {id_in {}} } {
 	    } else {
 		set default [PrefDefault $item]
 	    }
-	    if {([llength $default] > 1) && ([lindex $default 0] == "CHOICE")} {
-		PrefValueSet $varName [lindex $default 1]
-	    } else {
-		if {$default == "OFF"} {
-		    PrefValueSet $varName 0
-		} else {
-		    if {$default == "ON"} {
-			PrefValueSet $varName 1
-		    } else {
-			global PrefEntry
-			if [info exists PrefEntry($varName)] {
-			    set entry $PrefEntry($varName)
-			    $entry delete 0 end
-			    $entry insert 0 $default
-			}
-			PrefValueSet $varName $default
-		    }
+	    switch -regexp -- $default {
+		^ON$		{PrefValueSet $varName 1}
+		^OFF$		{PrefValueSet $varName 0}
+		"^CHOICE "	{PrefValueSet $varName [lindex $default 1]}
+		default		{
+				global PrefEntry
+				if [info exists PrefEntry($varName)] {
+				    set entry $PrefEntry($varName)
+				    $entry delete 0 end
+				    $entry insert 0 $default
+				}
+				PrefValueSet $varName $default
 		}
 	    }
 	}
     }
 }
+proc Preferences_Tweak { _varName } {
+    # Change a single setting in the preferences database.
+    # This assumes a preference value has been changes from outside the
+    # preferences UI.
+    global pref PrefEntry
+    set done 0
+    foreach id $pref(panes) {
+	foreach item $pref($id,prefs) {
+	    set varName [PrefVar $item]
+	    if {[string compare $varName $_varName] == 0} {
+		set xresName [PrefXres $item]
+		upvar #0 $varName x
+		if [info exists PrefEntry($varName)] {
+		    # Update the preferences user interface.
+		    $PrefEntry($varName) delete 0 end
+		    $PrefEntry($varName) insert 0 $x
+		}
+		set done 1
+		break
+	    }
+	}
+	if {$done} break
+    }
+    if {! $done} {
+	error "No resource associated with $_varName"
+    }
+    # Change one line
+    if [catch {open $pref(userDefaults) r} old] {
+	# No existing preferences, better save all of them
+	PreferencesSave nodismiss
+	return
+    }
+    if [catch {open $pref(userDefaults).new w} out] {
+	Exmh_Status "Cannot save in $pref(userDefaults).new: $out" warn
+	close $old
+	return
+    }
+    foreach line [split [read -nonewline $old] \n] {
+	if [regexp "^\\*$xresName:" $line] {
+	    puts $out "*$xresName: $x"
+	    set done 1
+	} else {
+	    puts $out $line
+	}
+    }
+    close $old
+    close $out
+    set new [glob $pref(userDefaults).new]
+    set old [file root $new]
+    if [catch {exec mv $new $old} err] {
+	Exmh_Status "Cannot install $new: $err"
+	return
+    }
+}
 proc Preferences_Resource { _varName _rname _default } {
     set _rval [option get . $_rname {}]
     if {$_rval != {}} {
@@ -517,3 +540,47 @@ proc Preferences_Resource { _varName _rname _default } {
 	PrefValueSet $_varName $_default
     }
 }
+
+proc PreferencesPaneHelp { id {gotoxres {}} } {
+    global pref tk_version
+    set ix [lsearch $pref(panes) $id]
+    set top .prefhelplong$ix
+    set t $top.t
+    set numLines 8
+
+    if [Exwin_Toplevel $top "Exmh '$id' Preferences Help" Help] {
+	wm group $top .pref$ix
+	Widget_AddBut $top.but help Help {PreferencesHelp}
+	Widget_Label  $top.but label {left fill} -text "Help for '$id' Preferences"
+	Widget_Text $top $numLines -setgrid true
+	$t tag configure headings -underline 1
+	$t insert end $id\n\n
+	$t tag add headings 1.0 "end -2c"
+	$t insert end $pref($id,text)\n\n\n
+	foreach item $pref($id,prefs) {
+	    PreferencesPaneHelpText $t $item
+	}
+	$t mark set help4_ 1.0
+	$t configure -state disabled
+    }
+
+    if {$tk_version >= 4.0} {
+	# make as much help text visible
+	$t see end
+        $t see help4_$gotoxres
+    } else {
+	$t yview end
+        $t yview help4_$gotoxres
+    }
+}
+proc PreferencesPaneHelpText { t item } {
+    set res [PrefXres $item] 
+    set var [PrefVar  $item] 
+    $t insert end [PrefComment $item]\n\n
+    $t tag add headings "insert -2 line" "insert -1c"
+    $t mark set help4_$res "insert -2 line"
+    foreach line [split [PrefHelp $item] \n] {
+	$t insert end \t$line\n
+    }
+    $t insert end "\n\ttk resource:  $res\n\ttcl variable: $var\n\n"
+}
