@@ -17,15 +17,18 @@ proc Widget_TextInit {} {
     # Preferences stuff moved to exwin.tcl
     # for the sake of the user interface.
 }
-proc Widget_TextInitText { t } {
+proc Widget_TextInitText {t} {
     global widgetText		;# Constrained scrolling module
     set widgetText($t,geo) {}
     global TextType		;# Text bindings module
     set TextType($t) text
-    selection handle $t [list Text_HandleSelRequest $t]
+    global tk_version
+    if {$tk_version < 4.0} {
+	selection handle $t [list Text_HandleSelRequest $t]
+    }
 
 }
-proc Widget_Text { frame height args } {
+proc Widget_Text {frame height args} {
     # Create the text widget used to display messages
     global exwin
     if ![info exists exwin(scrollbarSide)] {
@@ -33,9 +36,8 @@ proc Widget_Text { frame height args } {
     } else {
 	set side $exwin(scrollbarSide)
     }
-    set cmd [list text $frame.t -relief raised -bd 2 \
-		-yscroll [list WidgetScrollSet $frame.sv $frame.t] \
-		-wrap word]
+    set cmd [list text $frame.t -relief raised -bd 2  \
+		-yscroll [list WidgetScrollSet $frame.sv $frame.t]]
     if [catch [concat $cmd $args] t] {
 	puts stderr "Widget_Text (warning) $t"
 	set t [eval $cmd $args {-font fixed}]
@@ -58,60 +60,105 @@ proc Widget_Text { frame height args } {
 
     return $t
 }
-
-# For Keyboard scrolling
-
-proc Widget_TextPageOrNext { t } {
-    global widgetText
-    set bottom [lindex $widgetText($t,view) 3]
-    set end [expr [$t index end]-1]
-    if {$bottom >= $end} {
-	Ftoc_Next show implied
+proc Widget_TextPageOrNext {t} {
+    global widgetText tk_version
+    set next 0
+    if {$tk_version >= 4.0} {
+	set bottom [lindex [$t yview] 1]
+	set next [expr $bottom >= 1]
+    } else {
+	if [info exists widgetText($t,view)] {
+	    set bottom [lindex $widgetText($t,view) 3]
+	    set end [expr [$t index end]-1]
+	    set next [expr $bottom >= $end]
+	}
+    }
+    if {$next} {
+	Ftoc_NextImplied
     } else {
 	Widget_TextPageDown $t
     }
 }
-
-proc Widget_TextPageDown { t } {
-    global widgetText
-    Exmh_Debug WidgetTextPageDown $t
-    if [info exists widgetText($t,view)] {
-	set bottom [lindex $widgetText($t,view) 3]
-	WidgetTextYview $t [expr $bottom-$widgetText(scrollContext)]
+proc Widget_TextPageDown {t} {
+    global widgetText tk_version
+    if {$tk_version >= 4.0} {
+	WidgetTextYview $t scroll 1 pages
+    } else {
+	if [info exists widgetText($t,view)] {
+	    set bottom [lindex $widgetText($t,view) 3]
+	    WidgetTextYview $t [expr $bottom-$widgetText(scrollContext)]
+	}
+    }
+    $t mark set insert @1,1
+}
+proc Widget_TextPageUp {t} {
+    global widgetText tk_version
+    if {$tk_version >= 4.0} {
+	WidgetTextYview $t scroll -1 pages
+    } else {
+	if [info exists widgetText($t,view)] {
+	    set top [lindex $widgetText($t,view) 2]
+	    set hei [lindex $widgetText($t,view) 1]
+	    WidgetTextYview $t [expr $top-$hei+$widgetText(scrollContext)]
+	}
+    }
+    $t mark set insert @1,1
+}
+proc Widget_TextLineDown {t} {
+    global widgetText tk_version
+    if {$tk_version >= 4.0} {
+	$t yview scroll 1 units
+    } else {
+	if [info exists widgetText($t,view)] {
+	    set top [lindex $widgetText($t,view) 2]
+	    WidgetTextYview $t [expr $top+1]
+	}
     }
 }
-proc Widget_TextPageUp { t } {
-    global widgetText
-    Exmh_Debug WidgetTextPageUp $t
-    if [info exists widgetText($t,view)] {
-	set top [lindex $widgetText($t,view) 2]
-	set hei [lindex $widgetText($t,view) 1]
-	WidgetTextYview $t [expr $top-$hei+$widgetText(scrollContext)]
+proc Widget_TextLineUp {t} {
+    global widgetText tk_version
+    if {$tk_version >= 4.0} {
+	$t yview scroll -1 units
+    } else {
+	if [info exists widgetText($t,view)] {
+	    set top [lindex $widgetText($t,view) 2]
+	    WidgetTextYview $t [expr $top-1]
+	}
     }
 }
-
-proc Widget_TextLineDown { t } {
+proc WidgetTextYview4.0 {w args} {
     global widgetText
-    if [info exists widgetText($t,view)] {
-        set top [lindex $widgetText($t,view) 2]
-        WidgetTextYview $t [expr $top+1]
+    set dir [lindex $args 1]
+    set op  [lindex $args 0]
+    set units  [lindex $args 2]
+    set view [$w yview]
+    set next [expr 1 - [lindex $view 1]]
+    set prev [lindex $view 0]
+    set span [expr 1. - $next - $prev]
+    if {!$widgetText(smoothScroll) ||
+	($op != "scroll") ||
+	($units != "pages") ||
+        ($dir > 0 && $next >= $span) ||
+        ($dir < 0 && $prev >= $span)} {
+        eval $w yview $args
+    } else {
+        while {($dir > 0 && $next > 0) ||
+                ($dir < 0 && $prev > 0)} {
+            $w yview scroll [expr ($dir > 0) ? 2 : -2] unit
+            update idletasks
+            set view [$w yview]
+            set next [expr 1 - [lindex $view 1]]
+            set prev [lindex $view 0]
+        }
     }
 }
-
-proc Widget_TextLineUp { t } {
-    global widgetText
-    if [info exists widgetText($t,view)] {
-        set top [lindex $widgetText($t,view) 2]
-        WidgetTextYview $t [expr $top-1]
+proc WidgetTextYview {t args} {
+    global widgetText tk_version
+    if {$tk_version >= 4.0} {
+	return [eval WidgetTextYview4.0 $t $args]
     }
-}
-
-# Constrained scrolling
-
-proc WidgetTextYview { t args } {
-    global widgetText
     if {!$widgetText(constrained) &&
-	!($widgetText(constrainFtoc) && [string match *.ftoc.* $t])} {
+	 !($widgetText(constrainFtoc) && [string match *.ftoc.* $t])} {
 	eval {$t yview} $args
 	return
     }
@@ -154,7 +201,7 @@ proc WidgetTextYview { t args } {
     set mark [$t index [list $mark +1 lines]]
     $t yview $mark
 }
-proc WidgetScrollSet { s t args } {
+proc WidgetScrollSet {s t args} {
     global widgetText
     set widgetText($t,view) $args
     if {$s != {}} {
@@ -163,13 +210,13 @@ proc WidgetScrollSet { s t args } {
 	}
     }
 }
-proc WidgetTextMark { t y } {
+proc WidgetTextMark {t y} {
     global widgetText
 #    Exmh_Debug WidgetTextMark $t $y
     set widgetText($t,mark) $y			;# Remember mark point
     scan [$t index @1,1] %d widgetText($t,top)	;# and starting top line
 }
-proc WidgetTextDragto { t y speed } {
+proc WidgetTextDragto {t y speed} {
     global widgetText
     if ![info exists widgetText($t,mark)] {
 	return
@@ -197,7 +244,7 @@ proc WidgetTextDragto { t y speed } {
 	set widgetText($t,top) $newtopline
     }
 }
-proc WidgetTextGridY { t } {
+proc WidgetTextGridY {t} {
     global widgetText
     set geo [split [winfo geometry $t] +x]
     if ![info exists widgetText($t,geo)] {
@@ -213,10 +260,7 @@ proc WidgetTextGridY { t } {
     }
     return $widgetText($t,gridY)
 }
-
-# Text selection, with power-drag-scroll when mouse leaves the window
-
-proc WidgetTextSelBegin { w x y how } {
+proc WidgetTextSelBegin {w x y how} {
     global widgetText
     set widgetText($w,extend) 0
     switch -- $how {
@@ -225,13 +269,13 @@ proc WidgetTextSelBegin { w x y how } {
 	line {Text_LineSelect $w @$x,$y}
     }
 }
-proc WidgetTextSelAgain { w x y } {
+proc WidgetTextSelAgain {w x y} {
     global widgetText
     set widgetText($w,extend) 0
     tk_textResetAnchor $w @$x,$y
     Text_SelectTo $w @$x,$y
 }
-proc WidgetTextSelMotion { w x y } {
+proc WidgetTextSelMotion {w x y} {
     global widgetText
 
     if ![info exists widgetText($w,extend)] {
@@ -259,8 +303,7 @@ proc WidgetTextSelMotion { w x y } {
 	}
     }
 }
-
-proc WidgetTextSelExtend { w } {
+proc WidgetTextSelExtend {w} {
     global widgetText
 
     if {![info exists widgetText($w,extend)] ||
@@ -284,10 +327,47 @@ proc WidgetTextSelExtend { w } {
 	after $widgetText(selectDelay) [list WidgetTextSelExtend $w]
     }
 }
-
-proc WidgetTextSelDone { w } {
+proc WidgetTextSelDone {w} {
     global widgetText
     catch {unset widgetText($w,extend)}
     Text_SelectionEnd $w 1
 }
+proc Widget_TextEnd {w} {
+    global tk_version
+    scan [$w index end] %d i
+    if {$tk_version >= 4.0} {
+	incr i -2
+    } else {
+	incr i -1
+    }
+}
 
+if {$tk_version >= 4.0} {
+    proc tk_textResetAnchor {args} {eval tkTextResetAnchor $args}
+}
+# Fill out the text widget with enough blanks to allow the
+# given line to appear at the top.
+proc Widget_TextPad {w top} {
+    global tk_version
+    if {$tk_version < 4.0} {
+	return
+    }
+    # Assume -height is ok, even though can be wrong after resize
+    # set height [$w cget -height]
+    #
+    # The above replaced by the following, thanks to 
+    # Harvey Thompson <harveyt@sco.com> and
+    # John Haxby <J.Haxby@isode.com>
+
+    # Add newlines so that text in all window
+    while {[$w bbox "end -1c"] != ""} {
+	$w insert end \n
+    }
+    # Compute height even if using different fonts
+    set height [expr int([$w index @0,65535]) - int([$w index @0,0]) + 1] 
+    set last [$w index "end -1c"]
+    Exmh_Debug Widget_TextPad h=$height last=$last top=$top
+    for {} {$last - $top < $height+1} {set last [expr $last + 1.0]} {
+	$w insert end \n
+    }
+}
