@@ -98,44 +98,49 @@ proc Widget_AddButDef {par but {where {right padx 1}} } {
     # Create a Packed button.  Return the button pathname
     set cmd2 [list button $par.$but]
     if [catch $cmd2 t] {
-	puts stderr "Widget_AddBut (warning) $t"
+	catch {puts stderr "Widget_AddButDef (warning) $t"}
 	eval $cmd2 {-font fixed}
     }
     pack append $par $par.$but $where
     return $par.$but
 }
 proc Widget_ReEvalCmd { but } {
-    uplevel "$but config -command \[eval list \[lindex \[$but config -command] 4]]"
+    global tk_version
+    if {$tk_version >= 4.0} {
+	uplevel "$but config -command \[subst \[$but cget -command]]"
+    } else {
+	uplevel "$but config -command \[eval list \[lindex \[$but config -command] 4]]"
+    }
 }
 
 proc Widget_AddBut {par but txt cmd {where {right padx 1}} } {
     # Create a Packed button.  Return the button pathname
     set cmd2 [list button $par.$but -text $txt -command $cmd]
     if [catch $cmd2 t] {
-	puts stderr "Widget_AddBut (warning) $t"
+	catch {puts stderr "Widget_AddBut (warning) $t"}
 	eval $cmd2 {-font fixed}
     }
     pack append $par $par.$but $where
     return $par.$but
 }
 
-proc Widget_CheckBut {par but txt var {where {right padx 1}} } {
+proc Widget_CheckBut {par but txt var {where {right padx 1}} args} {
     # Create a check button.  Return the button pathname
     set cmd [list checkbutton $par.$but -text $txt -variable $var]
-    if [catch $cmd t] {
-	puts stderr "Widget_CheckBut (warning) $t"
-	eval $cmd {-font fixed}
+    if [catch [concat $cmd $args] t] {
+	catch {puts stderr "Widget_CheckBut (warning) $t"}
+	eval $cmd {-font fixed} $args
     }
     pack append $par $par.$but $where
     return $par.$but
 }
 
-proc Widget_RadioBut {par but txt var {where {right padx 1}} } {
+proc Widget_RadioBut {par but txt var {where {right padx 1}} args} {
     # Create a radio button.  Return the button pathname
     set cmd [list radiobutton $par.$but -text $txt -variable $var -value $txt]
-    if [catch $cmd t] {
-	puts stderr "Widget_RadioBut (warning) $t"
-	eval $cmd {-font fixed}
+    if [catch [concat $cmd $args] t] {
+	catch {puts stderr "Widget_RadioBut (warning) $t"}
+	eval $cmd {-font fixed} $args
     }
     pack append $par $par.$but $where
     return $par.$but
@@ -145,7 +150,7 @@ proc Widget_AddMenuBDef {par b {where {left fill}} } {
     # Create a button and a menu to go with it.  Return the menu pathname
     set cmd [list menubutton $par.$b -menu $par.$b.m]
     if [catch $cmd t] {
-	puts stderr "Widget_AddMenuB (warning) $t"
+	catch {puts stderr "Widget_AddMenuBDef (warning) $t"}
 	eval $cmd {-font fixed}
     }
     if [catch {menu $par.$b.m}] {
@@ -158,7 +163,7 @@ proc Widget_AddMenuB {par b {label {}} {where {left fill}} } {
     # Create a button and a menu to go with it.  Return the menu pathname
     set cmd [list menubutton $par.$b -menu $par.$b.m -text $label]
     if [catch $cmd t] {
-	puts stderr "Widget_AddMenuB (warning) $t"
+	catch {puts stderr "Widget_AddMenuB (warning) $t"}
 	eval $cmd {-font fixed}
     }
     if [catch {menu $par.$b.m}] {
@@ -172,7 +177,7 @@ proc Widget_AddMenuItem {m l cmd {accel NONE}} {
     # Create a menu command entry with optional accelerator string.
     set cmd2 [list $m add command -label $l  -command $cmd]
     if [catch $cmd2 t] {
-	puts stderr "Widget_AddMenuItem (warning) $t"
+	catch {puts stderr "Widget_AddMenuItem (warning) $t"}
 	eval $cmd2 {-font fixed}
     }
     if {$accel != "NONE"} {
@@ -192,31 +197,34 @@ proc Widget_RadioMenuItem {m l {cmd { }} {var {}} args} {
     }
     set cmd2 [list $m add radio -label $l  -variable $var -command $cmd]
     if [catch [concat $cmd2 $args] t] {
-	puts stderr "Widget_RadioMenuItem (warning) $t"
+	catch {puts stderr "Widget_RadioMenuItem (warning) $t"}
 	eval $cmd2 $args {-font fixed}
     }
 }
 
-proc Widget_CheckMenuItem {m l {c { }} {var {}}} {
+proc Widget_CheckMenuItem {m l {c { }} {var {}} args} {
     # Create a Check button menu entry.  By default all check entries
     # have their own variable.
     if {$var == {}} {
 	set var v$m.$l
     }
     set cmd [list $m add check -label $l -variable $var -command $c]
-    if [catch $cmd t] {
-	puts stderr "Widget_CheckMenuItem (warning) $t"
-	eval $cmd {-font fixed}
+    if [catch [concat $cmd $args] t] {
+	catch {puts stderr "Widget_CheckMenuItem (warning) $t"}
+	eval $cmd $args {-font fixed}
     }
     return $var
 }
 
-proc Widget_CascadeMenuItem {menu l {c { }} {submenu {}}} {
+proc Widget_CascadeMenuItem {menu l {c { }} {sub {}}} {
     # Create a cascade menu entry.
-    # Note that the menu parameter is really a menu
+    # Note that sub is the component after menu
+    if [catch {menu $menu.$sub} submenu] {
+	set submenu [menu $menu.$sub -font fixed]
+    }
     set cmd [list $menu add cascade -label $l -menu $submenu -command $c]
     if [catch $cmd t] {
-	puts stderr "Widget_CheckMenuItem (warning) $t"
+	catch {puts stderr "Widget_CascadeMenuItem (warning) $t"}
 	eval $cmd {-font fixed}
     }
     return $submenu
@@ -228,7 +236,7 @@ proc Widget_SimpleText { frame name {where {expand fill}} args } {
     global exwin
     set cmd [list text $frame.$name]
     if [catch [concat $cmd $args] t] {
-	puts stderr "Widget_TextLine (warning) $t"
+	catch {puts stderr "Widget_SimpleText (warning) $t"}
 	set t [eval $cmd $args {-font fixed}]
     }
     pack append $frame $t $where
@@ -243,7 +251,7 @@ proc Widget_SimpleText { frame name {where {expand fill}} args } {
 proc Widget_Message { frame {name msg} args} {
     set cmd [list message $frame.$name]
     if [catch [concat $cmd $args] t] {
-	puts stderr "Widget_Message (warning) $t"
+	catch {puts stderr "Widget_Message (warning) $t"}
 	eval $cmd $args {-font fixed}
     }
     pack append $frame $frame.$name {top fill expand}
@@ -252,7 +260,7 @@ proc Widget_Message { frame {name msg} args} {
 proc Widget_Label { frame {name label} {where {left fill}} args} {
     set cmd [list label $frame.$name ]
     if [catch [concat $cmd $args] t] {
-	puts stderr "Widget_Label (warning) $t"
+	catch {puts stderr "Widget_Label (warning) $t"}
 	eval $cmd $args {-font fixed}
     }
     pack append $frame $frame.$name $where
@@ -261,7 +269,7 @@ proc Widget_Label { frame {name label} {where {left fill}} args} {
 proc Widget_Entry { frame {name entry} {where {left fill}} args} {
     set cmd [list entry $frame.$name ]
     if [catch [concat $cmd $args] t] {
-	puts stderr "Widget_Entry (warning) $t"
+	catch {puts stderr "Widget_Entry (warning) $t"}
 	eval $cmd $args {-font fixed}
     }
     pack append $frame $frame.$name $where
@@ -278,51 +286,109 @@ proc Widget_ReadOnlyText { w } {
     }
 }
 proc Widget_PlaceDialog { parent frame } {
-    place $frame -in $parent -x 0 -y 0
-    update idletask
-    set W [winfo width $parent]
-    set H [winfo height $parent]
-    set w [winfo width $frame]
-    set h [winfo height $frame]
-    place $frame -in $parent -x [expr ($W-$w)/2] -y [expr ($H-$h)/2]
+    place $frame -in $parent -relx 0.5 -rely 0.5 -anchor center
 }
 #
 # Widget_BeginEntries, Widget_LabeledEntry, and Widget_EndEntries
 # are used to create a set of labeled-entry widgets.  The labels
 # line up and <Tab> takes the focus from one to the next.
 #
-proc Widget_BeginEntries { {lwidth 10} {ewidth 20} {okCmd {}} } {
+proc Widget_BeginEntries { {lwidth 10} {ewidth 20} {okCmd {}} {link {}}} {
     global widgetEntry
-    catch {unset widgetEntry(last)}
     set widgetEntry(lwidth) $lwidth
     set widgetEntry(ewidth) $ewidth
     set widgetEntry(okCmd) $okCmd
+    catch {unset widgetEntry(last)}
+    if {$link != {}} {
+	set widgetEntry(last) $link
+	set widgetEntry(first) [lindex [bind $link <Tab>] 1]
+	if {[string length [string trim $widgetEntry(first)]] == 0} {
+	    error "Widget_BeginEntries link=$link"
+	}
+    }
 }
 proc Widget_LabeledEntry { w name textvar} {
-    global widgetEntry
+    global widgetEntry tk_version
     set f [frame $w -class LabeledEntry]
     Widget_Label $f label {left} -text $name -width $widgetEntry(lwidth)
-    Widget_Entry $f entry {right fillx expand} \
+    Widget_Entry $f entry {left fillx} \
 	-width $widgetEntry(ewidth) -textvariable $textvar
     pack $f -side top -fill x
     if [info exists widgetEntry(last)] {
 	bind $widgetEntry(last) <Tab> [list focus $f.entry]
+	bind $f.entry <Shift-Tab> [list focus $widgetEntry(last)]
     } else {
 	set widgetEntry(first) $f.entry
     }
-    bind $f.entry <Return> $widgetEntry(okCmd)
+    Widget_BindEntryCmd $f.entry <Return>  $widgetEntry(okCmd)
     set widgetEntry(last) $f.entry
-#    upvar #0 v $textvar
-#    if ![info exists v] {
-#	set v {}
-#   }
+    return $f
+}
+proc Widget_BindEntryCmd {entry  sequence cmd} {
+    global tk_version
+    if {$tk_version < 4.0} {
+	bind $entry $sequence $cmd
+    } else {
+	bindtags $entry [list $entry Entry [winfo toplevel $entry] all]
+	bind $entry $sequence "$cmd ; break"
+    }
+}
+proc Widget_EntryEntry { w labelvar textvar} {
+    global widgetEntry
+    set f [frame $w -class LabeledEntry]
+    Widget_Label $f label {left} -text "xx" -width $widgetEntry(lwidth)
+    Widget_Entry $f elabel {left} -textvariable $labelvar \
+	-width [expr $widgetEntry(lwidth) - 1]
+    pack forget $f.elabel
+    place $f.elabel -in $f.label -anchor ne -relx 1.0 -y -1 
+    Widget_Entry $f entry {left fillx} \
+	-width $widgetEntry(ewidth) -textvariable $textvar
+    pack $f -side top -fill x
+    bind $f.elabel <Tab> [list focus $f.entry]
+    bind $f.entry <Shift-Tab> [list focus $f.elabel]
+    if [info exists widgetEntry(last)] {
+	bind $widgetEntry(last) <Tab> [list focus $f.elabel]
+	bind $f.elabel <Shift-Tab> [list focus $widgetEntry(last)]
+    } else {
+	set widgetEntry(first) $f.elabel
+    }
+    Widget_BindEntryCmd $f.entry <Return> $widgetEntry(okCmd)
+    Widget_BindEntryCmd $f.elabel <Return> $widgetEntry(okCmd)
+    set widgetEntry(last) $f.entry
+    return $f
+}
+proc Widget_LabeledEntryOr { f iter textvar} {
+    global widgetEntry
+    if {$iter > 2} {
+	set pe $f.entry[expr $iter -1]
+    } else {
+	set pe $f.entry
+    }
+    Widget_Entry $f entry$iter {top fillx} -textvariable $textvar \
+    	-width [lindex [$pe config -width] 4]
+    set me $f.entry$iter
+    Widget_Bindtags $me [list $me Entry [winfo toplevel $me]]
+    bind $me <Return> [bind $pe <Return>]
+    bind $me <Tab> [bind $pe <Tab>]
+    bind $pe <Tab> [list focus $me]
+    bind $me <Shift-Tab> [list focus $pe]
+    bind [lindex [bind $me <Tab>] 1] <Shift-Tab> [list focus $me]
+    focus $me
     return $f
 }
 proc Widget_EndEntries {} {
     global widgetEntry
     if [info exists widgetEntry(first)] {
 	bind $widgetEntry(last) <Tab> [list focus $widgetEntry(first)]
+	bind $widgetEntry(first) <Shift-Tab> [list focus $widgetEntry(last)]
 	focus $widgetEntry(first)
+	return $widgetEntry(last)
+    }
+}
+proc Widget_Bindtags { w tags } {
+    global tk_version
+    if {$tk_version >= 4.0} {
+	bindtags $w $tags
     }
 }
 # Widget_ListEditor
@@ -345,6 +411,7 @@ proc Widget_EndEntries {} {
 #
 
 proc Widget_ListEditor {frame title entryvar {insert {}} {change {}} {delete {}} {select {}}} {
+    global exwin
 
     FontWidget label $frame.label \
 	-text $title
@@ -359,7 +426,6 @@ proc Widget_ListEditor {frame title entryvar {insert {}} {change {}} {delete {}}
 	-relief {sunken}
 
     FontWidget entry $frame.entry \
-	-width 20 \
 	-textvariable $entryvar
 
     frame $frame.buttons
@@ -383,6 +449,10 @@ proc Widget_ListEditor {frame title entryvar {insert {}} {change {}} {delete {}}
 	-side left -padx 10
 
     # bindings
+    global tk_version
+    if {$tk_version >= 4.0} {
+	$frame.listbox config -selectmode browse
+    }
     bind $frame.listbox <Any-B1-Motion> \
 	[list WidgetListSelect %W %y $entryvar $select]
     bind $frame.listbox <Any-Button-1> \
@@ -394,7 +464,7 @@ proc Widget_ListEditor {frame title entryvar {insert {}} {change {}} {delete {}}
     pack $frame.label -side top -fill x
     pack $frame.buttons -side bottom
     pack $frame.entry -side bottom -pady 5 -fill x
-    pack $frame.scrollbar -side right -padx 8 -fill y
+    pack $frame.scrollbar -side $exwin(scrollbarSide) -padx 8 -fill y
     pack $frame.listbox -side top -expand 1 -fill both
 
     return $frame
@@ -408,7 +478,7 @@ proc FontWidget { args } {
 proc WidgetListSelect { w y varName selCmd } {
     upvar #0 $varName entryvar
     set i [$w nearest $y]
-    $w select from $i
+    Widget_ListboxSelect $w $i
     set entryvar [$w get $i]
     eval $selCmd
 }
@@ -420,14 +490,89 @@ proc Widget_ListSearch { frame } {
     }
     set l $frame.listbox
     set size [$l size]
+    Widget_ListboxClear $l
     for {set i 0} {$i < $size} {incr i} {
 	if {[string match ${str}* [$l get $i]]} {
-	    $l select from $i
+	    Widget_ListboxSelect $l $i
 	    Widget_ListboxYview $l $i
 	    return
 	}
     }
-    $l select clear
+}
+
+proc Widget_ListboxYview { list i } {
+    global tk_version
+    if {$tk_version < 4.0} {
+	set height [lindex [split [lindex [$list config -geometry] 4] x] 1]
+	$list yview [expr $i - $height/2]
+    } else {
+	$list see $i
+    }
+}
+proc Widget_ListboxSelect { list i } {
+    global tk_version
+    if {$tk_version < 4.0} {
+	$list select from $i
+    } else {
+	$list select set $i
+    }
+}
+proc Widget_ListboxClear { list } {
+    global tk_version
+    if {$tk_version < 4.0} {
+	$list select clear
+    } else {
+	$list select clear 0 end
+    }
 }
 
 
+#
+# Procedures hiding the configuration resource hierarchy
+# 
+proc Widget_GetButDef { f } {
+    WidgetGetResources $f buttonlist
+}
+proc Widget_GetMenuBDef { f } {
+    WidgetGetResources $f menulist
+}
+proc Widget_GetEntryDef { m } {
+    WidgetGetResources $m entrylist
+}
+proc Widget_GetButGrDef { f g } {
+    WidgetGetResources $f g_$g
+}
+proc Widget_GetMenuGrDef { f g } {
+    WidgetGetResources $f gm_$g
+}
+# Only system groups allowed (mentioned in buttons.tcl)
+proc Widget_GetGroupDef { f } {
+    option get $f groups {}
+}
+proc WidgetGetResources { w resname } {
+    set res	[option get $w $resname {}]
+    set lres    [option get $w l$resname {}]
+    set ures 	[option get $w u$resname {}]
+
+    set l-res	[option get $w l-$resname {}]
+    set u-res	[option get $w u-$resname {}]
+
+    set list [WidgetResListSubtract $res ${l-res}]
+    set list [concat $list $lres]
+    set list [WidgetResListSubtract $list ${u-res}]
+
+    return [concat $list $ures]
+}
+
+proc WidgetResListSubtract { orglist remlist } {
+    #Remove words in 'remlist' from 'orglist'
+    set newlist $orglist
+    foreach dele $remlist {
+    	set tmplist ""
+	foreach item $newlist {
+	    if {"x$item" != "x$dele"} { lappend  tmplist $item }
+	}
+	set newlist $tmplist
+    }
+    return $newlist
+}
